#
# Copyright 2018-2021 Lars Pastewka
#           2018-2021 Michael Röttger
#           2019-2020 Antoine Sanner
#           2019 Kai Haase
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
SurfaceTopography profile from file input
"""

import xml.etree.ElementTree as ElementTree
from zipfile import ZipFile
import numpy as np

from .common import CHANNEL_NAME_INFO_KEY
from .Reader import ReaderBase, ChannelInfo
from ..Exceptions import MetadataAlreadyFixedByFile
from ..UniformLineScanAndTopography import Topography


###

def binary(func):
    def func_wrapper(fobj, *args, **kwargs):
        close_file = False
        if not hasattr(fobj, 'read'):
            fobj = open(fobj, 'rb')
            close_file = True
        try:
            retvals = func(fobj, *args, **kwargs)
        finally:
            if close_file:
                fobj.close()
        return retvals

    return func_wrapper


def make_wrapped_reader(reader_func, class_name='WrappedReader', format=None,
                        name=None, description=None):
    class WrappedReader(ReaderBase):
        """
        emulates the new implementation of the readers
        """

        _format = format
        _name = name
        _description = description

        def __init__(self, fobj):
            self._fobj = fobj
            self._file_position = 0
            if hasattr(fobj, 'tell'):
                self._file_position = fobj.tell()
            self._topography = reader_func(fobj)
            if CHANNEL_NAME_INFO_KEY in self._topography.info:
                self._channel_name = self._topography.info[CHANNEL_NAME_INFO_KEY]
                del self._topography.info[CHANNEL_NAME_INFO_KEY]
            else:
                self._channel_name = "Default"

        @property
        def channels(self):
            try:
                height_scale_factor = self._topography.height_scale_factor
            except AttributeError:
                height_scale_factor = None
                # None means: Not available in file

            return [ChannelInfo(
                self, 0,
                name=self._channel_name,
                dim=self._topography.dim,
                unit=self._topography.unit,
                uniform=self._topography.is_uniform,
                undefined_data=self._topography.has_undefined_data,
                info=self._topography.info,
                nb_grid_pts=self._topography.nb_grid_pts,
                physical_sizes=self._topography.physical_sizes,
                height_scale_factor=height_scale_factor)]

        def topography(self, channel_index=None, physical_sizes=None,
                       height_scale_factor=None, unit=None, info={}, periodic=False,
                       subdomain_locations=None, nb_subdomain_grid_pts=None):
            if channel_index is None:
                channel_index = self._default_channel_index

            if subdomain_locations is not None or nb_subdomain_grid_pts is not None:
                raise RuntimeError(
                    'This reader does not support MPI parallelization.')

            if channel_index != 0:
                raise RuntimeError('Reader supports only a single channel 0.')

            physical_sizes = self._check_physical_sizes(physical_sizes, self._topography.physical_sizes)

            # Rewind to position where the data is. Otherwise this method
            # cannot be called twice.
            if hasattr(self._fobj, 'seek'):
                self._fobj.seek(self._file_position)

            # Read again, but this time with physical_sizes and unit set (if not
            # specified in file)
            reader_kwargs = dict(height_scale_factor=height_scale_factor,
                                 unit=unit, info=info.copy(), periodic=periodic)
            if self._topography.physical_sizes is None:
                # file does not have physical sizes
                reader_kwargs['physical_sizes'] = physical_sizes
                # otherwise we won't add the argument, because that is not allowed any more

            return reader_func(self._fobj, **reader_kwargs)

        channels.__doc__ = ReaderBase.channels.__doc__
        topography.__doc__ = ReaderBase.topography.__doc__

    WrappedReader.__name__ = class_name
    return WrappedReader


def read_x3p(fobj, physical_sizes=None, height_scale_factor=None, unit=None, info={},
             periodic=False):
    """
    Load x3p-file.
    See: http://opengps.eu

    FIXME: Descriptive error messages. Probably needs to be made more robust.

    Keyword Arguments:
    fobj -- filename or file object
    """

    # Data types of binary container
    # See: https://sourceforge.net/p/open-gps/mwiki/X3p/
    dtype_map = {'I': np.dtype('<u2'),
                 'L': np.dtype('<u4'),
                 'F': np.dtype('f4'),
                 'D': np.dtype('f8')}

    with ZipFile(fobj, 'r') as x3p:
        xmlroot = ElementTree.parse(x3p.open('main.xml')).getroot()
        record1 = xmlroot.find('Record1')
        record3 = xmlroot.find('Record3')

        if record1 is None:
            raise IOError("'Record1' not found in XML.")
        if record3 is None:
            raise IOError("'Record3' not found in XML.")

        # Parse record1

        feature_type = record1.find('FeatureType')
        if feature_type.text != 'SUR':
            raise ValueError("FeatureType must be 'SUR'.")
        axes = record1.find('Axes')
        cx = axes.find('CX')
        cy = axes.find('CY')
        cz = axes.find('CZ')

        if cx.find('AxisType').text != 'I':
            raise ValueError(
                "CX AxisType is not 'I'. Don't know how to handle "
                "this.")
        if cy.find('AxisType').text != 'I':
            raise ValueError(
                "CY AxisType is not 'I'. Don't know how to handle "
                "this.")
        if cz.find('AxisType').text != 'A':
            raise ValueError(
                "CZ AxisType is not 'A'. Don't know how to handle "
                "this.")

        xinc = float(cx.find('Increment').text)
        yinc = float(cy.find('Increment').text)

        datatype = cz.find('DataType').text
        dtype = dtype_map[datatype]

        # Parse record3
        matrix_dimension = record3.find('MatrixDimension')
        nx = int(matrix_dimension.find('SizeX').text)
        ny = int(matrix_dimension.find('SizeY').text)
        nz = int(matrix_dimension.find('SizeZ').text)

        if nz != 1:
            raise ValueError('Z dimension has extend != 1. Volumetric data '
                             'is not supported.')

        data_link = record3.find('DataLink')
        binfn = data_link.find('PointDataLink').text

        rawdata = x3p.open(binfn).read(nx * ny * dtype.itemsize)
        data = np.frombuffer(rawdata, count=nx * ny * nz, dtype=dtype).reshape(nx, ny).T

    if physical_sizes is not None:
        raise MetadataAlreadyFixedByFile('physical_sizes')
    t = Topography(data, (xinc * nx, yinc * ny), unit=unit, info=info, periodic=periodic)
    if height_scale_factor is not None:
        t = t.scale(height_scale_factor)
    return t


X3PReader = make_wrapped_reader(
    read_x3p, class_name="X3PReader", format='x3p',
    name='XML 3D surface profile (X3P)', description='''
X3P is a container format conforming to the ISO 5436-2 (Geometrical Product
Specifications — Surface texture) standard. The format is defined in ISO
25178 and is a standardized format for the exchange of surface topography
data. The full specification of the format can be found
[here](http://www.opengps.eu/).
''')


@binary
def read_hgt(fobj, physical_sizes=None, height_scale_factor=None, unit=None, info={},
             periodic=False):
    """
    Read Shuttle Radar SurfaceTopography Mission (SRTM) topography data
    (.hgt extension).

    Keyword Arguments:
    fobj -- filename or file object
    """
    fobj.seek(0, 2)
    fsize = fobj.tell()
    fobj.seek(0)

    dim = int(np.sqrt(fsize / 2))
    if dim * dim * 2 != fsize:
        raise RuntimeError(
            'File physical_sizes of {0} bytes does not match file '
            'physical_sizes for a map of dimension {1}x{1}.'.format(fsize, dim))
    data = np.fromfile(fobj, dtype=np.dtype('>i2'),
                       count=dim * dim).reshape((dim, dim))

    if physical_sizes is None:
        topography = Topography(data, physical_sizes=data.shape, unit=unit, info=info, periodic=periodic)
    else:
        topography = Topography(data, physical_sizes=physical_sizes, unit=unit, info=info, periodic=periodic)
    if height_scale_factor is not None:
        topography = topography.scale(height_scale_factor)
    return topography


HGTReader = make_wrapped_reader(
    read_hgt, class_name="HGTReader", format='hgt',
    name='NASA shuttle radar topography mission', description='''
Data format of the NASA shuttle radar topography mission that recorded the '
earths topography. More information can be found
[here](https://www2.jpl.nasa.gov/srtm/).
                                ''')



#
# Copyright 2015-2016, 2018-2021 Lars Pastewka
#           2018-2019 Antoine Sanner
#           2015-2016 Till Junge
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Functions computing scalar roughness parameters
"""

import numpy as np

from ..Exceptions import ReentrantDataError
from ..HeightContainer import NonuniformLineScanInterface


def rms_height(topography):
    r"""
    Computes root-mean square height fluctuation of the line scan:

    .. math::

        h_\text{rms} = \left[\frac{1}{L} \int_0^L dx\, h^2(x)\right]^{1/2}

    Function approximates topography between data points as piece-wise linear.
    The piece-wise linear section between point :math:`i` and point
    :math:`i+1` contributes

    .. math::

        \int_{0}^{\Delta x_i} dx\, \left( h_i + \frac{\Delta h_i}{\Delta x_i} x \right)^2 = \frac{1}{3} \left( h_i^2 + h_{i+1}^2 + h_i h_{i+1} \right) \Delta x_i

    to the above integral, where :math:`\Delta x_i=x_{i+1}-x_i` and :math:`\Delta h_i=h_{i+1}-h_i`.

    Parameters
    ----------
    topography : :obj:`NonuniformLineScan`
        SurfaceTopography object containing height information.

    Returns
    -------
    rms_height : float or array
        Root-mean square height.
    """  # noqa: E501
    x, h = topography.positions_and_heights()
    dx = np.diff(x)
    if len(x) <= 1:
        return 0.0
    L = x[-1] - x[0]
    h0 = h - topography.mean()
    return np.sqrt(
        np.sum((h0[:-1] ** 2 + h0[1:] ** 2 + h0[:-1] * h0[1:]) * dx) / (3 * L))


def rms_slope(topography):
    r"""
    Computes root-mean square slope fluctuation of the line scan:

    .. math:: h^\prime_\text{rms} = \left[ \frac{1}{L} \int_0^L dx\, \left(\frac{\partial h}{\partial x}\right)^2 \right]^{1/2}

    Function approximates topography between data points as piece-wise linear.
    The piece-wise linear section between point :math:`i` and point
    :math:`i+1` contributes

    .. math:: \int_{0}^{\Delta x_i} dx\, \left( \frac{\Delta h_i}{\Delta x_i} \right)^2 = \frac{\Delta h_i^2}{\Delta x_i}

    where :math:`\Delta x_i=x_{i+1}-x_i` and :math:`\Delta h_i=h_{i+1}-h_i` to the above integral.

    Parameters
    ----------
    topography : :obj:`NonuniformLineScan`
        SurfaceTopography object containing height information.

    Returns
    -------
    rms_slope : float
        Root-mean square slope.
    """  # noqa: E501
    x, h = topography.positions_and_heights()
    dh = np.diff(h)
    dx = np.diff(x)

    if np.min(dx) <= 0:
        raise ReentrantDataError('This topography is reentrant (i.e. it contains overhangs). The rms slope '
                                 'cannot be computed for reentrant topographies.')

    L = x[-1] - x[0]

    return np.sqrt(np.sum(dh ** 2 / dx) / L)


def rms_curvature(topography):
    r"""
    Computes root-mean square slope fluctuation of the line scan:

    Parameters
    ----------
    topography : :obj:`NonuniformLineScan`
        SurfaceTopography object containing height information.

    Returns
    -------
    rms_slope : float
        Root-mean square slope.
    """
    x = topography.positions()
    d2 = topography.derivative(n=2)
    # The second derivative cannot be evaluated on the two end points
    L = x[-2] - x[1]

    return np.sqrt(np.trapz(d2 ** 2, x[1:-1]) / L)


# Register analysis functions from this module
NonuniformLineScanInterface.register_function('rms_height_from_profile', rms_height)
NonuniformLineScanInterface.register_function('rms_slope_from_profile', rms_slope)
NonuniformLineScanInterface.register_function('rms_curvature_from_profile', rms_curvature)



#
# Copyright 2018-2021 Lars Pastewka
#           2019 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Height-difference autocorrelation functions
"""

import numpy as np

from ..Exceptions import NoReliableDataError, UndefinedDataError
from ..HeightContainer import UniformTopographyInterface
from ..Support import doi
from ..Support.Regression import resample, resample_radial


@doi('10.1016/j.triboint.2018.02.002')
def autocorrelation_from_profile(self, reliable=True, resampling_method='bin-average', collocation='log',
                                 nb_points=None, nb_points_per_decade=10):
    r"""
    Compute the one-dimensional height-difference autocorrelation function
    (ACF).

    For non-periodic surfaces the ACF at distance d is given by:

       .. math::
         :nowrap:

         \begin{equation}
         \begin{split}

          \text{ACF}(d) =& \sum_{i=0}^{n-d-1} \frac{1}{n-d} \frac{1}{2} \left( h_i - h_{i+d} \right)^2 \\

                        =& \frac{1}{2(n-d)} \sum_{i=0}^{n-d-1} \left( h_i^2 + h_{i+d}^2 \right)
                         - \frac{1}{n-d} \sum_{i=0}^{n-d-1} h_i h_{i+d}
          \end{split}
          \end{equation}


    Parameters
    ----------
    self : :obj:`SurfaceTopography` or :obj:`UniformLineScan`
        Container storing the uniform topography map
    reliable : bool, optional
        Only return data deemed reliable. (Default: True)
    resampling_method : str, optional
        Method can be None for no resampling (return on the grid of the
        data) 'bin-average' for simple bin averaging and 'gaussian-process'
        for Gaussian process regression.
        (Default: 'bin-average')
    collocation : {'log', 'quadratic', 'linear', array_like}, optional
        Resampling grid. Specifying 'log' yields collocation points
        equally spaced on a log scale, 'quadratic' yields bins with
        similar number of data points and 'linear' yields linear bins.
        Alternatively, it is possible to explicitly specify the bin edges.
        If bin_edges are explicitly specified, then `rmax` and `nbins` is
        ignored. (Default: 'log')
    nb_points : int, optional
        Number of bins for averaging. Bins are automatically determined if set
        to None. (Default: None)
    nb_points_per_decade : int, optional
        Number of points per decade for log-spaced collocation points.
        (Default: None)

    Returns
    -------
    r : array
        Distances. (Units: length)
    A : array
        Autocorrelation function. (Units: length**2)
    """  # noqa: E501
    if self.has_undefined_data:
        raise UndefinedDataError('This topography has undefined data (missing data points). Autocorrelation cannot be '
                                 'computed for topographies with missing data points.')

    try:
        nx, ny = self.nb_grid_pts
        sx, sy = self.physical_sizes
    except ValueError:
        nx, = self.nb_grid_pts
        sx, = self.physical_sizes

    p = self.heights()

    if self.is_periodic:
        # Compute height-height autocorrelation function from a convolution
        # using FFT. This is periodic by nature.
        surface_qy = np.fft.fft(p, axis=0)
        C_qy = abs(surface_qy) ** 2  # pylint: disable=invalid-name
        A_xy = np.fft.ifft(C_qy, axis=0).real / nx

        # Convert height-height autocorrelation to height-difference
        # autocorrelation:
        #     <(h(x) - h(x+d))^2>/2 = <h^2(x)> - <h(x)h(x+d)>
        A_xy = A_xy[0] - A_xy

        A = A_xy[:nx // 2]
        A[1:nx // 2] += A_xy[nx - 1:(nx + 1) // 2:-1]
        A /= 2

        r = sx * np.arange(nx // 2) / nx
        max_distance = sx / 2
    else:
        # Compute height-height autocorrelation function. We need to zero
        # pad the FFT for the nonperiodic case in order to separate images.
        surface_qy = np.fft.fft(p, n=2 * nx - 1, axis=0)
        C_qy = abs(surface_qy) ** 2  # pylint: disable=invalid-name
        A_xy = np.fft.ifft(C_qy, axis=0).real

        # Correction to turn height-height into height-difference
        # autocorrelation:
        #     <(h(x) - h(x+d))^2>_d/2 = <h^2(x)>_d - <h(x)h(x+d)>_d
        # but we need to take care about h_rms^2=<h^2(x)>, which in the
        # nonperiodic case needs to be computed only over a subsection of
        # the surface. This is because the average < >_d now depends on d,
        # which determines the number of data points that are actually
        # included into the computation of <h(x)h(x+d)>_d. h_rms^2 needs to
        # be computed over the same data points.
        p_sq = p ** 2
        A0_xy = (p_sq.cumsum(axis=0)[::-1] + p_sq[::-1].cumsum(axis=0)[::-1]) / 2

        # Convert height-height autocorrelation to height-difference
        # autocorrelation
        A = ((A0_xy - A_xy[:nx]).T / (nx - np.arange(nx))).T

        r = sx * np.arange(nx) / nx
        max_distance = sx

    # The factor of two comes from the fact that the short cutoff is estimated
    # from the curvature but the ACF is the slope, see 10.1016/j.apsadv.2021.100190
    if resampling_method is None:
        short_cutoff = self.short_reliability_cutoff() if reliable else None
        if short_cutoff is not None:
            mask = r > short_cutoff / 2
            if mask.sum() == 0:
                raise NoReliableDataError('Dataset contains no reliable data.')
            r = r[mask]
            A = A[mask]
        if self.dim == 2:
            return r, A.mean(axis=1)
        else:
            return r, A
    else:
        short_cutoff = self.short_reliability_cutoff(2 * sx / nx) if reliable else 2 * sx / nx
        if collocation == 'log':
            # Exclude zero distance because that does not work on a log-scale
            r = r[1:]
            A = A[1:]
        if self.dim == 2:
            r = np.resize(r, (A.shape[1], r.shape[0])).T.ravel()
            A = np.ravel(A)
        r, _, A, _ = resample(r, A, min_value=short_cutoff / 2, max_value=max_distance, collocation=collocation,
                              nb_points=nb_points, nb_points_per_decade=nb_points_per_decade, method=resampling_method)
        return r, A


@doi('10.1016/j.triboint.2018.02.002')
def autocorrelation_from_area(self, reliable=True, collocation='log', nb_points=None, nb_points_per_decade=5,
                              return_map=False, resampling_method='bin-average'):
    """
    Compute height-difference autocorrelation function and radial average.

    Parameters
    ----------
    self : :obj:`SurfaceTopography`
        Container storing the (two-dimensional) topography map.
    reliable : bool, optional
        Only return data deemed reliable. (Default: True)
    collocation : {'log', 'quadratic', 'linear', array_like}, optional
        Resampling grid. Specifying 'log' yields collocation points
        equally spaced on a log scale, 'quadratic' yields bins with
        similar number of data points and 'linear' yields linear bins.
        Alternatively, it is possible to explicitly specify the bin edges.
        If bin_edges are explicitly specified, then `rmax` and `nbins` is
        ignored. (Default: 'log')
    nb_points : int, optional
        Number of bins for averaging. Bins are automatically determined if set
        to None. (Default: None)
    nb_points_per_decade : int, optional
        Number of points per decade for log-spaced collocation points.
        (Default: None)
    return_map : bool, optional
        Return full 2D autocorrelation map. (Default: False)
    resampling_method : str, optional
        Method can be 'bin-average' for simple bin averaging and
        'gaussian-process' for Gaussian process regression.
        (Default: 'bin-average')

    Returns
    -------
    r : array
        Distances. (Units: length)
    A : array
        Radially averaged autocorrelation function. (Units: length**2)
    A_xy : array
        2D autocorrelation function. Only returned if return_map=True.
        (Units: length**2)
    """
    if self.has_undefined_data:
        raise UndefinedDataError('This topography has undefined data (missing data points). Autocorrelation cannot be '
                                 'computed for topographies with missing data points.')

    nx, ny = self.nb_grid_pts
    sx, sy = self.physical_sizes

    # The factor of two comes from the fact that the short cutoff is estimated
    # from the curvature but the ACF is the slope, see 10.1016/j.apsadv.2021.100190
    short_cutoff = self.short_reliability_cutoff(np.mean(self.pixel_size)) if reliable else np.mean(self.pixel_size)

    # Compute FFT and normalize
    if self.is_periodic:
        surface_qk = np.fft.fft2(self[...])
        C_qk = abs(surface_qk) ** 2  # pylint: disable=invalid-name
        A_xy = np.fft.ifft2(C_qk).real / (nx * ny)

        # Convert height-height autocorrelation to height-difference
        # autocorrelation
        A_xy = A_xy[0, 0] - A_xy

        # Radial average
        r_val, r_edges, A_val, _ = resample_radial(A_xy, physical_sizes=(sx, sy), nb_points=nb_points,
                                                   nb_points_per_decade=nb_points_per_decade, collocation=collocation,
                                                   min_radius=short_cutoff, max_radius=(sx + sy) / 4,
                                                   method=resampling_method)
    else:
        p = self.heights()

        # Compute height-height autocorrelation function
        surface_qk = np.fft.fft2(p, s=(2 * nx - 1, 2 * ny - 1))
        C_qk = abs(surface_qk) ** 2  # pylint: disable=invalid-name
        A_xy = np.fft.ifft2(C_qk).real

        # Correction to turn height-height into height-difference
        # autocorrelation
        p_sq = p ** 2
        A0_xy = (p_sq.cumsum(axis=0).cumsum(axis=1)[::-1, ::-1] +
                 p_sq[::-1, ::-1].cumsum(axis=0).cumsum(axis=1)[::-1, ::-1]) / 2

        # Convert height-height autocorrelation to height-difference
        # autocorrelation
        A_xy = (A0_xy - A_xy[:nx, :ny]) / (
                (nx - np.arange(nx)).reshape(-1, 1) * (ny - np.arange(ny)).reshape(1, -1))

        # Radial average
        r_val, r_edges, A_val, _ = resample_radial(A_xy, physical_sizes=(sx, sy), collocation=collocation,
                                                   nb_points=nb_points, nb_points_per_decade=nb_points_per_decade,
                                                   min_radius=short_cutoff, max_radius=(sx + sy) / 2, full=False,
                                                   method=resampling_method)

    if return_map:
        return r_val, A_val, A_xy
    else:
        return r_val, A_val


# Register analysis functions from this module
UniformTopographyInterface.register_function('autocorrelation_from_profile', autocorrelation_from_profile)
UniformTopographyInterface.register_function('autocorrelation_from_area', autocorrelation_from_area)



#
# Copyright 2022 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from struct import calcsize, unpack


def decode(stream_obj, structure_format, byte_order='@'):
    """
    Decode a binary stream given the sequence of binary entries. Strings are
    stripped of zeros and white spaces.

    Parameters
    ----------
    stream_obj : stream-like object
        Binary stream to decode.
    structure_format : list of tuples
        List of tuples describing the sequence of entries in the binary
        stream. Each tuple consists of two entries
            (name, format)
        that give the name of the entry and the format. We support the format
        defined in the `struct` module, plus 'u' for UTF-8 and 'U' for UTF-16.
    byte_order : str
        Byte order (see `struct.unpack`).

    Returns
    -------
    data : dict
        Dictionary with decoded data entries.
    """
    def mogrify_format(format):
        return_format = format
        if format.endswith('b'):  # bytes
            return_format = format[:-1] + 's'
        elif format.endswith('u'):  # UTF-8
            return_format = format[:-1] + 's'
        elif format.endswith('U'):  # UTF-16
            return_format = str(2*int(format[:-1])) + 's'
        return return_format

    def decode_data(data, format):
        if format.endswith('s'):
            return data.decode('latin1').strip('\x00').strip(' ')
        elif format.endswith('u'):
            return data.decode('utf-8').strip('\x00').strip(' ')
        elif format.endswith('U'):
            return data.decode('utf-16').strip('\x00').strip(' ')
        else:
            return data

    format_str = byte_order + ''.join([mogrify_format(format) for name, format in structure_format])
    unpacked_data = unpack(format_str, stream_obj.read(calcsize(format_str)))
    return {name: decode_data(data, format) for (name, format), data in zip(structure_format, unpacked_data)}



#
# Copyright 2019-2021 Lars Pastewka
#           2019 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Utility function for conversion between uniform and nonuniform
representations.
"""

import numpy as np

from SurfaceTopography.HeightContainer import NonuniformLineScanInterface
from SurfaceTopography.UniformLineScanAndTopography import UniformLineScan, DecoratedUniformTopography


class UniformlyInterpolatedLineScan(DecoratedUniformTopography):
    """
    Interpolate a topography onto a uniform grid.
    """

    def __init__(self, topography, nb_points=None, padding=0, nb_interpolate=None, pixel_size=None, info={}):
        """
        Convert a nonuniform line scan to a uniform line scan by explicit
        linear interpolation between the discrete coordinates of the
        nonuniform scan.

        `nb_points`, `nb_interpolate` or `pixel_size` needs to be specified.

        Parameters
        ----------
        topography : :obj:`SurfaceTopography` or :obj:`UniformLineScan`
            SurfaceTopography to interpolate.
        nb_points : int, optional
            Number of equidistant grid points. Number of points will be
            automatically computed from the mean grid spacing and
            `nb_interpolate` if set to None. (Default: None)
        padding : int, optional
            Number of padding grid points, zeros appended to the data.
            (Default: 0)
        nb_interpolate : int, optional
            Number of grid points to between closest points on surface.
            (Default: None)
        pixel_size : float, optional
            Grid spacing. (Default: None)
        """
        super().__init__(topography, info=info)
        self.nb_points = nb_points
        self.nb_interpolate = nb_interpolate
        self.padding = padding
        self.in_pixel_size = pixel_size
        self._update_nb_points_and_pixel_size()

        # This is populated with functions from the nonuniform topography, but
        # this is a uniform topography
        self._functions = UniformLineScan._functions

    def _update_nb_points_and_pixel_size(self):
        """Automatically compute `nb_points` and `pixel_size` if it is None"""
        s, = self.parent_topography.physical_sizes
        x = self.parent_topography.positions()
        min_dist = np.min(np.diff(x))
        if min_dist <= 0:
            raise RuntimeError('This is a reentrant nonuniform line scan. Reentrant line scans cannot be converted '
                               'to uniform line scans.')

        if self.in_pixel_size is not None:
            if self.nb_points is not None or self.nb_interpolate is not None:
                raise ValueError('You need to specify either `nb_points`, `nb_interpolate` or `pixel_size`.')
            self._nb_points = int(s / self.in_pixel_size) + 1
            self._pixel_size = self.in_pixel_size
        elif self.nb_interpolate is not None:
            if self.nb_points is not None or self.in_pixel_size is not None:
                raise ValueError('You need to specify either `nb_points`, `nb_interpolate` or `pixel_size`.')
            self._nb_points = self.nb_interpolate * int(s / min_dist)
            self._pixel_size = s / (self._nb_points - 1)
        elif self.nb_points is not None:
            if self.nb_interpolate is not None or self.in_pixel_size is not None:
                raise ValueError('You need to specify either `nb_points`, `nb_interpolate` or `pixel_size`.')
            self._nb_points = self.nb_points
            self._pixel_size = s / (self._nb_points - 1)
        else:
            raise ValueError('You need to specify one of `nb_points`, `nb_interpolate` or `pixel_size`.')

    def __getstate__(self):
        """ is called and the returned object is pickled as the contents for
            the instance
        """
        state = super().__getstate__(), self.nb_points, self.padding, self.nb_interpolate, self.in_pixel_size
        return state

    def __setstate__(self, state):
        """ Upon unpickling, it is called with the unpickled state
        Keyword Arguments:
        state -- result of __getstate__
        """
        superstate, self.nb_points, self.padding, self.nb_interpolate, self.in_pixel_size = state
        super().__setstate__(superstate)
        self._update_nb_points_and_pixel_size()

    # Implement abstract methods of AbstractHeightContainer

    @property
    def dim(self):
        return 1

    @property
    def physical_sizes(self):
        return (self._nb_points + self.padding - 1) * self._pixel_size,

    @property
    def is_periodic(self):
        return False

    @property
    def is_uniform(self):
        return True

    # Implement uniform line scan interface

    @property
    def nb_grid_pts(self):
        """Return nb_grid_pts, i.e. number of pixels, of the topography."""
        return self._nb_points + self.padding,

    @property
    def pixel_size(self):
        return self._pixel_size,

    @property
    def area_per_pt(self):
        return self._pixel_size

    @property
    def has_undefined_data(self):
        return False

    def positions(self):
        return np.arange(self._nb_points + self.padding) * self._pixel_size

    def heights(self):
        """ Computes the rescaled profile.
        """
        x = self.positions()
        return np.interp(x, *self.parent_topography.positions_and_heights())


# Register pipeline functions from this module
NonuniformLineScanInterface.register_function('to_uniform', UniformlyInterpolatedLineScan)



#
# Copyright 2020-2021 Lars Pastewka
#           2019-2020 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import logging
from collections import defaultdict

import numpy as np

from ..Exceptions import NoReliableDataError
from .SurfaceContainer import SurfaceContainer

_log = logging.Logger(__name__)


def scale_dependent_statistical_property(self, func, n, unit, nb_points_per_decade=10, distances=None,
                                         interpolation='linear', reliable=True, progress_callback=None, threshold=4):
    """
    Compute statistical properties of a topography container (i.e. of a set of
    topographies and line scans) at specific scales. These properties are
    statistics of derivatives carried out at specific scales, as computed
    using the `derivative` pipeline function of the topography and/or line
    scan.

    The specific statistical properties is computed by the `func` argument.
    The output of `func` needs to be homogeneous, i.e. if an array is returned,
    this array must have the same size independent of the derivative data that
    is fed into `func`. The results of `func` are averaged over all topographies
    and/or line scans with each topography/line scan weighted equally. (The
    underlying assumption is that a topography/line scan represents a
    statistically independent dataset.)

    All topographies/line scans are converted to the same unit given by the
    `unit` argument before analysis.

    Parameters
    ----------
    self : SurfaceContainer
        Container object containing the underlying data sets.
    func : callable
        The function that computes the statistical properties:

            ``func(dx, dy=None) -> np.ndarray``

        A function taking the derivative in x-direction and optionally the
        derivative in y-direction (only for topographies, i.e. maps). The
        function needs to be able to ignore the second argument as a container
        can be a mixture of topographies and line scans. The function can
        return a scalar value or an array, but the array size must be fixed.
    n : int
        Order of derivative.
    unit : str
        Unit of the distance array. All topographies are converted to this
        unit before the derivative is computed.
    nb_points_per_decade : int, optional
        Number of points per decade in length for automatic grid construction.
        (Default: 10)
    distances : float or np.ndarray, optional
        Characteristic distances at which the derivatives are computed. If
        this is an array, then the statistical property is computed at each
        of these distances. (Default: None)
    interpolation : str, optional
        Interpolation method to use for computing derivatives at distances
        that do not equal an integer multiple of the grid spacing. Use
        'linear' for a local liner interpolation or 'fourier' for global
        Fourier interpolation. Note that Fourier interpolation carries large
        errors for nonperiodic topographies and should be used with care.
        (Default: 'linear')
    reliable : bool, optional
        Only incorporate data deemed reliable. (Default: True)
    progress_callback : func, optional
        Function taking iteration and the total number of iterations as
        arguments for progress reporting. (Default: None)
    threshold : int, optional
        Defines the minimal amount of data points of the probability distribution
        to calculate the statistical properties with func and returns a np.nan if
        the value is below the threshold.
        E.g. the scipy.stats.kstat function needs at least 4 data points to
        calculate the 4th cumulant function, otherwise it returns nan or inf
        (Default: 4)

    Returns
    -------
    statistical_fingerprint : np.ndarray or np.ma.masked_array
        Array containing the result of `func`, average over all topographies
        and line scans. If multiple distances are provided, then this is a
        masked array that contains results for each distance. If no data
        exists for a certain distance, then the mask is set to true for that
        distance.

    Examples
    --------
    This example yields the the height-difference autocorrelation function in
    the x-direction, i.e. it should give a result identical to the average of
    `autocorrelation_from_profile` executed for all topographies/line scans in
    the container.

    Note that the statistics function (the lambda expression below) needs to be
    able to accept one argument (for line scans) and two arguments
    (for topographies).

    >>> s = c.scale_dependent_statistical_property(lambda x, y=None: np.var(x), n=1, distance=[0.1, 1.0, 10], unit='um')
    """
    results = defaultdict(list)
    for i, topography in enumerate(self):
        if progress_callback is not None:
            progress_callback(i, len(self))

        topography = topography.to_unit(unit)

        # Only compute the statistical property for distances that actually exist for this specific topography...
        lower, upper = topography.bandwidth()
        # ...and that are reliable
        if reliable:
            short_cutoff = topography.short_reliability_cutoff()
            if short_cutoff is None:
                short_cutoff = 0
            lower = max(short_cutoff, lower)

        if distances is None:
            if lower >= upper:
                _log.warning(f'Topography {topography} contributes no data to average.')
                continue

            # The automatic grid must include the full decades, i.e. 0.01, 0.1, 1, 10, 100, etc. as values. This
            # ensures that the grid of subsequent calculations are aligned.
            lower_decade, upper_decade = int(np.floor(np.log10(lower))), int(np.ceil(np.log10(upper)))
            existing_distances = np.logspace(lower_decade, upper_decade,
                                             (upper_decade - lower_decade) * nb_points_per_decade + 1)
            unique_distance_index = lower_decade * nb_points_per_decade + np.arange(len(existing_distances))
        else:
            existing_distances = np.array(distances)
            unique_distance_index = np.arange(len(distances))
        # For the factor n see 10.1016/j.apsadv.2021.100190
        m = np.logical_and(existing_distances > n * lower, existing_distances < upper)
        existing_distances = existing_distances[m]
        unique_distance_index = unique_distance_index[m]

        # Are there any distances left?
        if len(existing_distances) > 0:
            # Yes! Let's compute the statistical properties at these scales
            _, stat = topography.scale_dependent_statistical_property(
                func, n=n, distance=existing_distances, interpolation=interpolation, threshold=threshold)
            # Append results to our return values
            for i, e, s in zip(unique_distance_index, existing_distances, stat):
                results[i] += [(e, s)]
        else:
            _log.warning(f'Topography {topography} contributes no data to average.')

    if progress_callback is not None:
        progress_callback(len(self), len(self))

    if len(results) == 0:
        raise NoReliableDataError('Container contains no reliable data.')

    if distances is not None:
        # If distances are specified by the user, we return exactly those distances; whether data actually exists is
        # indicated through the mask of a masked array
        return distances, np.ma.masked_array(
            [np.nanmean(results[i], axis=0)[1] if i in results else np.nan for i in range(len(distances))],
            mask=np.array([i not in results for i in range(len(distances))])
        )
    else:
        # If distances are not specified by the user, the distance array contains only distances where data exists
        sorted_results = sorted(results.items(), key=lambda x: x[0])
        distances, properties = np.array([np.nanmean(vals, axis=0) for i, vals in sorted_results]).T
        return distances, properties


SurfaceContainer.register_function('scale_dependent_statistical_property', scale_dependent_statistical_property)



#
# Copyright 2019-2021 Lars Pastewka
#           2020-2021 Michael Röttger
#           2019 Kai Haase
#           2019 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os

from ..Exceptions import CannotDetectFileFormat, CorruptFile, MetadataAlreadyFixedByFile, ReadFileError, \
    UnknownFileFormatGiven  # noqa: F401

# Old-style readers
from .FromFile import HGTReader, X3PReader
from .Text import AscReader, XYZReader

# New-style readers
from .AL3D import AL3DReader
from .DI import DIReader
from .H5 import H5Reader
from .IBW import IBWReader
from .Matlab import MatReader
from .Mitutoyo import MitutoyoReader
from .MI import MIReader
from .NC import NCReader
from .NPY import NPYReader
from .PS import PSReader
from .OPD import OPDReader
from .OPDx import OPDxReader
from .SUR import SURReader
from .VK import VKReader
from .ZON import ZONReader

# Only writers
import SurfaceTopography.IO.DZI  # noqa: F401

from .Reader import ReaderBase  # noqa: F401

readers = [
    # XYZ must come before ASC, because 2D XYZ is a specialized ASC
    XYZReader,
    AscReader,
    DIReader,
    MatReader,
    OPDReader,
    OPDxReader,
    X3PReader,
    IBWReader,
    MIReader,
    MitutoyoReader,
    NCReader,
    # NCReader must come before H5Reader, because NC4 *is* a specialized form of HDF5
    H5Reader,
    NPYReader,
    PSReader,
    SURReader,
    VKReader,
    ZONReader,
    AL3DReader,
    # HGT reader should come last as there is no file magic
    HGTReader,
]

lookup_reader_by_format = {}
for reader in readers:
    lookup_reader_by_format[reader.format()] = reader


def detect_format(fobj, comm=None):
    """
    Detect file format based on its content.

    Arguments
    ---------
    fobj : filename or file object
    comm : mpi communicator, optional
    """
    msg = ""
    for reader in readers:
        try:
            if comm is not None:
                reader(fobj, comm)
            else:
                reader(fobj)
            return reader.format()
        except Exception as err:
            msg += "tried {}: \n {}\n\n".format(reader.__name__, err)
        finally:
            if hasattr(fobj, 'seek'):
                # if the reader crashes the cursor in the file-like object
                # has to be set back to the top of the file
                fobj.seek(0)
    raise CannotDetectFileFormat(msg)


def open_topography(fobj, format=None, communicator=None):
    r"""
    Returns a reader object for the file `fobj`. The reader interface mirrors
    the topography interface and can be used to extract meta data (number of
    grid points, physical sizes, etc.) without reading the full topography in
    memory.

    Parameters
    ----------
    fobj : str or filelike object
        path of the file or filelike object
    format : str, optional
        specify in which format the file should be interpreted
    communicator : mpi4py or NuMPI communicator object
        MPI communicator handling inter-process communication

    Returns
    -------
    Instance of a :class:`SurfaceTopography.IO.Reader.ReaderBase` subclass
    according to the format.

    Examples
    --------
    Simplest read workflow:

    >>> reader = open_topography("filename")
    >>> top = reader.topography()

    The first topography in file is returned, independently of whether
    the file has multiple channels or not.

    You can always check the channels and their metadata with
    `reader.channels()`. This returns a list of dicts containing attributes `name`,
    `physical_sizes`, `nb_grid_pts`, ``unit` and `height_scale_factor:

    >>> reader.channels
    [{'name': 'ZSensor',
      'nb_grid_pts': (256, 256),
      'physical_sizes': (9999.999999999998, 9999.999999999998),
      'unit': 'nm',
      'height_scale_factor': 0.29638271279074097},
     {'name': 'AmplitudeError',
      'nb_grid_pts': (256, 256),
      'physical_sizes': (10.0, 10.0),
      'unit': ('µm', None),
      'height_scale_factor': 0.04577566528320313}]

    Here the channel 'ZSensor' offers a topography with sizes of
    10000 nm in each dimension.

    You can choose it by giving the index 0 to channel
    (you would use `channel=1` for the second):

    >>> top = reader.topography(channel=0)

    The returned topography has the physical sizes found in the file.

    You can also prescribe some attributes when reading the topography:

    >>> top = reader.topography(channel=0, physical_sizes=(10.,10.), info={"unit":"µm"})

    In order to plot the topography with matplotlib, you can use

    >>> plt.pcolormesh(*top.positions_and_heights())

    with origin in the lower left and correct tick labels at x and y axes, or

    >>> plt.imshow(top.heights().T)

    with origin in the upper left (inverted y axis).
    """  # noqa: E501
    if communicator is not None:
        kwargs = {"communicator": communicator}
    else:
        kwargs = {}

    if not hasattr(fobj, 'read'):  # fobj is a path
        if not os.path.isfile(fobj):
            raise FileExistsError("file {} not found".format(fobj))

    if format is None:
        msg = ""
        for reader in readers:
            try:
                return reader(fobj, **kwargs)
            except Exception as err:
                msg += "tried {}: \n {}\n\n".format(reader.__name__, err)
            finally:
                if hasattr(fobj, 'seek'):
                    # if the reader crashes the cursor in the file-like object
                    # has to be set back to the top of the file
                    fobj.seek(0)
        raise CannotDetectFileFormat(msg)
    else:
        if format not in lookup_reader_by_format.keys():
            raise UnknownFileFormatGiven(
                "{} not in registered file formats {}".format(
                    fobj, lookup_reader_by_format.keys()))
        return lookup_reader_by_format[format](fobj, **kwargs)


def read_topography(fn, format=None, communicator=None, **kwargs):
    r"""
    Returns a topography object representing the topograpgy in the file `fobj`.
    If there are multiple data channels within this file, the default channel
    is returned. The default channel depends on the file format; see
    documentation of the respective reader on this.

    Parameters
    ----------
    fobj : str or filelike object
        path of the file or filelike object
    format : str, optional
        specify in which format the file should be interpreted
    communicator : mpi4py or NuMPI communicator object
        MPI communicator handling inter-process communication
    channel : int
        Number of the channel to load. See also `channels` method.
    physical_sizes : tuple of floats
        Physical size of the topography. It is necessary to specify this
        if no physical size is found in the data file. If there is a
        physical size in the file, then specifying this parameter will raise
        an exception.
    height_scale_factor : float
        Can be used to set height scale factor if not found in the data file.
    info : dict
        This dictionary will be appended to the info dictionary returned
        by the reader.
    periodic: bool
        Wether the SurfaceTopography should be interpreted as one period of a
        periodic surface. This will affect the PSD and autocorrelation
        calculations (windowing)
    subdomain_locations : tuple of ints
        Origin (location) of the subdomain handled by the present MPI process.
    nb_subdomain_grid_pts : tuple of ints
        Number of grid points within the subdomain handled by the present
        MPI process.

    Returns
    -------
    topography : subclass of :obj:`HeightContainer`
        The object containing the actual topography data.

    Raises
    ------
    MetadataAlreadyDefined
        Raised if given arguments for `physical_sizes` or `height_scale_factor`
        although it's already given in the file.
    """
    with open_topography(fn, format=format,
                         communicator=communicator) as reader:
        t = reader.topography(**kwargs)
    return t



# pop the current directory from search path
# python interpreter adds this to a top level script
# but we will likely have a name conflict (runtests.py .vs runtests package)
import sys; #sys.path.pop(0)
from runtests.mpi import Tester
import os

tester = Tester(os.path.join(os.path.abspath(__file__)), "SurfaceTopography")

tester.main(sys.argv[1:])


#
# Copyright 2019-2021 Lars Pastewka
#           2020-2021 Michael Röttger
#           2019 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

#
# The DI file format is described in detail here:
# http://www.physics.arizona.edu/~smanne/DI/software/fileformats.html
#

import re
from datetime import datetime

import numpy as np

from ..Exceptions import CorruptFile, MetadataAlreadyFixedByFile
from ..UniformLineScanAndTopography import Topography
from ..Support.UnitConversion import get_unit_conversion_factor, length_units, mangle_length_unit_utf8

from .Reader import ReaderBase, ChannelInfo


###

class DIReader(ReaderBase):
    _format = 'di'
    _name = 'Bruker Dimension, Veeco Nanoscope, Digital Instruments Nanoscope'
    _description = '''
Digital Instruments Nanoscope (also Veeco Nanoscope and Bruker Dimension)
files typically have a three-digit number as the file extension (.001, .002, .003, ...).
Newer versions of this file format have the extension .spm. This format contains
information on the physical size of the topography map as well as its units.
The reader supports V4.3 and later version of the format.
'''

    def __init__(self, fobj):
        """
        Load Digital Instrument's Nanoscope files.

        Arguments
        ---------
        fobj : filename or file object
             File or data stream to open.
        """
        self._fobj = fobj
        close_file = False
        if not hasattr(fobj, 'read'):
            fobj = open(fobj, 'rb')
            close_file = True
        try:
            parameters = []
            section_name = None
            section_dict = {}

            L = fobj.readline().decode('latin-1').strip()
            while L and L.lower() != r'\*file list end':
                if L.startswith('\\*'):
                    if section_name is not None:
                        parameters += [(section_name, section_dict)]
                    new_section_name = L[2:].lower()
                    if section_name is None:
                        if new_section_name != 'file list':
                            raise IOError("Header must start with the "
                                          "'File list' section.")
                    section_name = new_section_name
                    section_dict = {}
                elif L.startswith('\\'):
                    if section_name is None:
                        raise IOError('Encountered key before section '
                                      'header.')
                    s = L[1:].split(': ', 1)
                    try:
                        key, value = s
                    except ValueError:
                        key, = s
                        value = ''
                    section_dict[key.lower()] = value.strip()
                else:
                    raise IOError(f"Header line '{L}' does not start with a slash.")
                L = fobj.readline().decode('latin-1').strip()
            if section_name is None:
                raise IOError('No sections found in header.')
            parameters += [(section_name, section_dict)]

            self._channels = []
            self._offsets = []

            scanner = {}
            equipment = {}
            info = {}
            for n, p in parameters:
                if n == 'file list':
                    if 'date' in p:
                        info['acquisition_time'] = str(datetime.strptime(p['date'], '%I:%M:%S %p %a %b %d %Y'))
                elif n == 'scanner list' or n == 'ciao scan list':
                    scanner.update(p)
                elif n == 'equipment list':
                    equipment.update(p)
                elif n == 'ciao image list':
                    image_data_key = re.match(r'^S \[(.*?)\] ',
                                              p['@2:image data']).group(1)

                    nx = int(p['samps/line'])
                    ny = int(p['number of lines'])

                    s = p['scan size'].split(' ', 2)
                    sx = float(s[0])
                    sy = float(s[1])

                    xy_unit = mangle_length_unit_utf8(s[2])
                    offset = int(p['data offset'])
                    self._offsets.append(offset)

                    length = int(p['data length'])
                    elsize = int(p['bytes/pixel'])
                    binary_scale = 1
                    info['bytes_per_pixel'] = elsize
                    if elsize == 4:
                        binary_scale = 1 / 65536  # Rescale 32-bit integer to a 16-bit range
                    elif elsize != 2:
                        raise IOError(f"Don't know how to handle {elsize} bytes per pixel data.")
                    if nx * ny * elsize != length:
                        raise IOError(f'File reports a data block of length {length}, but computing the size of the '
                                      f'data block from the number of grid points and the per-pixel storage yields '
                                      f'a value of {nx * ny * elsize}.')

                    scale_re = re.match(
                        r'^V \[(.*?)\] \(([0-9\.]+) (.*)\/LSB\) (.*) '
                        r'(.*)', p['@2:z scale'])
                    quantity = scale_re.group(1).lower()
                    hard_scale = float(scale_re.group(4)) / 65536
                    hard_unit = scale_re.group(5)

                    s = scanner['@' + quantity].split()
                    if s[0] != 'V' or len(s) < 2:
                        raise IOError('Malformed Nanoscope DI file.')
                    soft_scale = float(s[1])

                    height_unit = None
                    hard_to_soft = 1.0
                    if len(s) > 2:
                        # Check units
                        height_unit, soft_unit = s[2].split('/')
                        hard_to_soft = get_unit_conversion_factor(hard_unit,
                                                                  soft_unit)
                        if hard_to_soft is None:
                            raise RuntimeError(
                                "Units for hard (={}) and soft (={}) "
                                "scale differ for '{}'. Don't know how "
                                "to handle this.".format(hard_unit,
                                                         soft_unit,
                                                         image_data_key))
                    if height_unit in length_units:
                        height_unit = mangle_length_unit_utf8(height_unit)
                        if xy_unit != height_unit:
                            fac = get_unit_conversion_factor(xy_unit,
                                                             height_unit)
                            sx *= fac
                            sy *= fac
                            xy_unit = height_unit
                        unit = height_unit
                    else:
                        unit = (xy_unit, height_unit)

                    height_scale_factor = hard_scale * hard_to_soft * soft_scale * binary_scale

                    if 'microscope' in equipment:
                        info['instrument'] = {'name': equipment['microscope']}
                    elif 'description' in equipment:
                        info['instrument'] = {'name': equipment['description']}

                    channel = ChannelInfo(self,
                                          len(self._channels),
                                          name=image_data_key,
                                          dim=2,
                                          nb_grid_pts=(nx, ny),
                                          physical_sizes=(sx, sy),
                                          height_scale_factor=height_scale_factor,
                                          periodic=False,
                                          uniform=True,
                                          unit=unit,
                                          info=info)
                    self._channels.append(channel)

            # Check that all channels can be read
            fobj.seek(0, 2)
            file_size = fobj.tell()
            for channel in self._channels:
                offset = self._offsets[channel.index]
                # We seek to the end of the data buffer, this should not raise an exception
                nx, ny = channel.nb_grid_pts
                elsize = channel.info['bytes_per_pixel']
                if offset + nx * ny * elsize > file_size:
                    raise CorruptFile('File is not large enough to contain all data buffers.')

        finally:
            if close_file:
                fobj.close()

    @property
    def channels(self):
        return self._channels

    def topography(self, channel_index=None, physical_sizes=None,
                   height_scale_factor=None, unit=None, info={}, periodic=False,
                   subdomain_locations=None, nb_subdomain_grid_pts=None):

        if channel_index is None:
            channel_index = self._default_channel_index

        if subdomain_locations is not None or \
                nb_subdomain_grid_pts is not None:
            raise RuntimeError(
                'This reader does not support MPI parallelization.')
        close_file = False
        if not hasattr(self._fobj, 'read'):
            fobj = open(self._fobj, 'rb')
            close_file = True
        else:
            fobj = self._fobj

        channel = self._channels[channel_index]

        if unit is not None:
            raise MetadataAlreadyFixedByFile('unit')

        sx, sy = self._check_physical_sizes(physical_sizes,
                                            channel.physical_sizes)

        nx, ny = channel.nb_grid_pts

        offset = self._offsets[channel_index]
        if channel.info['bytes_per_pixel'] == 2:
            dtype = np.dtype('<i2')
        elif channel.info['bytes_per_pixel'] == 4:
            dtype = np.dtype('<i4')
        else:
            raise IOError(f"Don't know how to handle {info['bytes_per_pixel']} bytes per pixel data.")

        assert channel.info['bytes_per_pixel'] == dtype.itemsize

        ###################################

        fobj.seek(offset)
        rawdata = fobj.read(nx * ny * dtype.itemsize)
        unscaleddata = np.frombuffer(rawdata, count=nx * ny, dtype=dtype).reshape(nx, ny)

        # internal information from file
        _info = dict(data_source=channel.name)
        _info.update(info)
        if 'acquisition_time' in channel.info:
            _info['acquisition_time'] = channel.info['acquisition_time']
        if 'instrument' in channel.info:
            try:
                # This can be a nested dictionary!
                _info['instrument'].update(channel.info['instrument'])
            except KeyError:
                _info['instrument'] = channel.info['instrument']

        # it is not allowed to provide extra `physical_sizes` here:
        if physical_sizes is not None:
            raise MetadataAlreadyFixedByFile('physical_sizes')

        # the orientation of the heights is modified in order to match
        # the image of gwyddion when plotted with imshow(t.heights().T)
        # or pcolormesh(t.heights().T) for origin in lower left and
        # with inverted y axis (cartesian coordinate system)

        surface = Topography(np.fliplr(unscaleddata.T), physical_sizes=(sx, sy), unit=channel.unit, info=_info,
                             periodic=periodic)
        if height_scale_factor is None:
            height_scale_factor = channel.height_scale_factor
        elif channel.height_scale_factor is not None:
            raise MetadataAlreadyFixedByFile('height_scale_factor')
        if height_scale_factor is not None:
            surface = surface.scale(height_scale_factor)

        if close_file:
            fobj.close()

        return surface

    channels.__doc__ = ReaderBase.channels.__doc__
    topography.__doc__ = ReaderBase.topography.__doc__



#
# Copyright 2020 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import muFFT

import numpy as np


def make_fft(topography, fft='mpi', communicator=None):
    """
    Instantiate a muFFT object that can compute the Fourier transform of the
    topography and has the same decomposition layout (or raise an error if
    this is not the case).

    Parameters
    ----------
    topography : :obj:`SurfaceTopography`
        Container storing the topography map.
    communicator, optional : mpi4py communicator or NuMPI stub communicator
        Communicator object. Use communicator from topography object if not
        present. (Default: None)
    """

    # We only initialize this once and attach it to the topography object
    if hasattr(topography, '_mufft'):
        return topography._mufft

    if topography.is_domain_decomposed:
        fft = muFFT.FFT(topography.nb_grid_pts, fft=fft,
                        communicator=topography.communicator if communicator is None else communicator)
        if fft.subdomain_locations != topography.subdomain_locations or \
                fft.nb_subdomain_grid_pts != topography.nb_subdomain_grid_pts:
            raise RuntimeError('muFFT suggested a domain decomposition that '
                               'differs from the decomposition of the topography.')
    else:
        fft = muFFT.FFT(topography.nb_grid_pts)
    topography._mufft = fft
    return fft


def get_window_2D(window, nx, ny, physical_sizes=None):
    """
    Construct a rotationally symmetric window for windowing two-dimensional
    signals (fields).

    Parameters
    ----------
    window : str or np.ndarray
        If window is an np.ndarray, it is just passed through (i.e.
        immediately returned) from this function. If window is a string,
        then a corresponding window will be constructed. Currently the only
        support window is 'hann'.
    nx : int
        Number of grid points of the signal in x-direction
    ny : int
        Number of grid points of the signal in y-direction
    physical_sizes : tuple of float
        The physical size of the underlying signal. If None, then the number
        of grid points nx, ny are used as the physical size. This parameter
        is used to adjust the aspect ratio of the signal. (Default: None)

    Returns
    -------
    window : np.ndarray
        Numerical window, multiply the signal with this window
    """

    if isinstance(window, np.ndarray):
        if window.shape != (nx, ny):
            raise TypeError('Window shape (= {0}x{1}) must match the number of points of the signal (={2}x{3})'
                            .format(*window.shape, nx, ny))
        return window

    if physical_sizes is None:
        sx, sy = nx, ny
    else:
        sx, sy = physical_sizes
    if window == 'hann':
        maxr = min(sx, sy) / 2
        r = np.sqrt((sx * (np.arange(nx).reshape(-1, 1) - nx // 2) / nx) ** 2 +
                    (sy * (np.arange(ny).reshape(1, -1) - ny // 2) / ny) ** 2)
        win = 0.5 + 0.5 * np.cos(np.pi * r / maxr)
        win[r > maxr] = 0.0
        return win
    else:
        raise ValueError("Unknown window type '{}'".format(window))



#
# Copyright 2018-2021 Lars Pastewka
#           2018-2021 Michael Röttger
#           2019-2020 Antoine Sanner
#           2019 Kai Haase
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import re

import numpy as np

from ..Exceptions import MetadataAlreadyFixedByFile
from ..HeightContainer import UniformTopographyInterface
from ..NonuniformLineScan import NonuniformLineScan
from ..UniformLineScanAndTopography import Topography, UniformLineScan
from ..Support.UnitConversion import length_units, mangle_length_unit_utf8
from .common import CHANNEL_NAME_INFO_KEY, text
from .FromFile import make_wrapped_reader

###


@text
def read_matrix(fobj, physical_sizes=None, unit=None, height_scale_factor=None, periodic=False):
    """
    Reads a surface profile from a text file and presents in in a
    SurfaceTopography-conformant manner. No additional parsing of
    meta-information is carried out.

    Keyword Arguments:
    fobj -- filename or file object
    """
    arr = np.loadtxt(fobj)
    if physical_sizes is None:
        surface = Topography(arr, arr.shape, periodic=periodic, unit=unit)
    else:
        surface = Topography(arr, physical_sizes, periodic=periodic, unit=unit)
    if height_scale_factor is not None:
        surface = surface.scale(height_scale_factor)
    return surface


MatrixReader = make_wrapped_reader(
    read_matrix, class_name="MatrixReader", format='matrix',
    name='Plain text (matrix)')


@text
def read_asc(fobj, physical_sizes=None, height_scale_factor=None, x_factor=1.0,
             z_factor=None, unit=None, info={}, periodic=False):
    # pylint: disable=too-many-branches,too-many-statements,invalid-name
    """
    Reads a surface profile (topography) from an generic asc file and presents
    it in a surface-conformant manner. Applies some heuristic to extract
    meta-information for different file formats.

    The info dict of the topography returned is a copy from the given info dict
    and may have some extra keys added:
    - "unit": a common unit for the data, for dimensions and heights
    - "channel_name": the name of the channel (if unknown "Default" is used")

    Parameters
    ----------

    fobj: filename or file object
    physical_sizes: tuple or None
    height_scale_factor: float or None
    x_factor: float
        multiplication factor for physical_sizes
    z_factor: float or None
        multiplication factor for height
    info: dict
    periodic: bool
    """
    _float_regex = r'[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?'

    checks = list()
    # Resolution keywords
    checks.append((re.compile(r"\b(?:x-pixels|h)\b\s*=\s*([0-9]+)"), int,
                   "yres"))
    checks.append((re.compile(r"\b(?:y-pixels|w)\b\s*=\s*([0-9]+)"), int,
                   "xres"))

    # Size keywords
    checks.append(
        (re.compile(r"\b(?:x-length|Width|Breite)\b\s*(?:=|\:)\s*(?P<value>" +
                    _float_regex + ")(?P<unit>.*)"), float, "xsiz"))
    checks.append(
        (re.compile(r"\b(?:y-length|Height|Höhe)\b\s*(?:=|\:)\s*(?P<value>" +
                    _float_regex + ")(?P<unit>.*)"), float, "ysiz"))

    # Unit keywords
    checks.append(
        (re.compile(r"\b(?:x-unit)\b\s*(?:=|\:)\s*(\w+)"), str, "xunit"))
    checks.append(
        (re.compile(r"\b(?:y-unit)\b\s*(?:=|\:)\s*(\w+)"), str, "yunit"))
    checks.append(
        (re.compile(r"\b(?:z-unit|Value units)\b\s*(?:=|\:)\s*(\w+)"),
         str, "zunit"))

    # Scale factor keywords
    checks.append(
        (re.compile(r"(?:pixel\s+size)\s*=\s*(?P<value>" + _float_regex +
                    ")(?P<unit>.*)"), float, "xfac"))
    checks.append((
        re.compile((
                r"(?:height\s+conversion\s+factor\s+\(->\s+(?P<unit>.*)\))\s*="
                r"\s*(?P<value>" + _float_regex + ")")),
        float, "zfac"))
    # Channel name keywords
    checks.append((
        re.compile(r"\b(?:Channel|Kanal)\b\s*(?:=|\:)\s*([\w|\s]+)"),
        str, "channel_name"))

    xres = yres = xsiz = ysiz = xunit = yunit = zunit = xfac = yfac = None
    zfac = None
    channel_name = "Default"

    def process_comment(line):
        """Find and interpret known comments in the header of the asc file"""
        nonlocal xres, yres, xsiz, ysiz, xunit, yunit, zunit, data, xfac, yfac
        nonlocal zfac
        nonlocal channel_name
        # TODO Is 'fun' needed here? No used so far..
        for reg, fun, key in checks:
            match = reg.search(line)
            if match is None:
                continue
            if key == 'xres':
                xres = int(match.group(1))
            elif key == 'yres':
                yres = int(match.group(1))
            elif key == 'xsiz':
                xsiz = float(match.group('value'))
                x = match.group('unit')
                if x:
                    xunit = mangle_length_unit_utf8(x)
            elif key == 'ysiz':
                ysiz = float(match.group('value'))
                y = match.group('unit')
                if y:
                    yunit = mangle_length_unit_utf8(y)
            elif key == 'xunit':
                xunit = mangle_length_unit_utf8(match.group(1))
            elif key == 'yunit':
                yunit = mangle_length_unit_utf8(match.group(1))
            elif key == 'zunit':
                zunit = mangle_length_unit_utf8(match.group(1))
            elif key == 'xfac':
                xfac = float(match.group('value'))
                xunit = mangle_length_unit_utf8(match.group('unit'))
            elif key == 'zfac':
                zfac = float(match.group('value'))
                zunit = mangle_length_unit_utf8(match.group('unit'))
            elif key == 'channel_name':
                channel_name = match.group(1).strip()

    data = []
    for line in fobj:
        line_elements = line.strip().split()
        if len(line_elements) > 0:
            try:
                data += [[float(strval) for strval in line_elements]]
            except ValueError:
                process_comment(line)

    if (height_scale_factor is not None) and (zfac is not None):
        # it should not be allowed to override a height scale factor if
        # there is one already in the file
        raise MetadataAlreadyFixedByFile('height_scale_factor')

    data = np.array(data).T
    nx, ny = data.shape
    if nx == 2 or ny == 2:
        raise Exception(
            "This file has just two rows or two columns and is more likely a "
            "line scan than a map.")
    if xres is not None and xres != nx:
        raise Exception(
            "The number of rows (={}) open_topography from the file '{}' "
            "does not match the nb_grid_pts in the file's metadata (={}).".format(nx, fobj, xres))
    if yres is not None and yres != ny:
        raise Exception(
            "The number of columns (={}) open_topography from the file '{}' "
            "does not match the nb_grid_pts in the file's metadata "
            "(={}).".format(ny, fobj, yres))

    # Handle scale factors
    if xfac is not None and yfac is None:
        yfac = xfac
    elif xfac is None and yfac is not None:
        xfac = yfac
    if xfac is not None:
        if xsiz is None:
            xsiz = xfac * nx
        else:
            xsiz *= xfac
    if yfac is not None:
        if ysiz is None:
            ysiz = yfac * ny
        else:
            ysiz *= yfac
    if z_factor is not None:
        zfac = z_factor if zfac is None else zfac * z_factor

    info = info.copy()

    # Handle units -> convert to target unit
    if xunit is None and zunit is not None:
        xunit = zunit
    if yunit is None and zunit is not None:
        yunit = zunit

    if unit is None:
        unit = zunit
    if unit is not None:
        if xunit is not None:
            xsiz *= length_units[xunit] / length_units[unit]
        if yunit is not None:
            ysiz *= length_units[yunit] / length_units[unit]
        if zunit is not None:
            if zfac is None:
                zfac = length_units[zunit] / length_units[unit]
            else:
                zfac *= length_units[zunit] / length_units[unit]

    # handle channel name
    # we use the info dict here to transfer the channel name
    info[CHANNEL_NAME_INFO_KEY] = channel_name

    # calculate physical sizes and generate topography
    if xsiz is not None and ysiz is not None:
        if physical_sizes is None:
            physical_sizes = (x_factor * xsiz, x_factor * ysiz)
        else:
            raise MetadataAlreadyFixedByFile('physical_sizes')
    if data.shape[0] == 1:
        if physical_sizes is not None and len(physical_sizes) > 1:
            physical_sizes = physical_sizes[0]
        surface = UniformLineScan(data[0, :], physical_sizes, unit=unit, info=info, periodic=periodic)
    else:
        surface = Topography(data, physical_sizes, unit=unit, info=info, periodic=periodic)
    if height_scale_factor is not None:
        zfac = height_scale_factor
    if zfac is not None and zfac != 1:
        surface = surface.scale(zfac)
    return surface


AscReader = make_wrapped_reader(read_asc, class_name="AscReader", format='asc',
                                name='Plain text (with headers)',
                                description='''
SurfaceTopography data stored in plain text (ASCII) format needs to be stored
in a matrix format. Each row contains the height information for subsequent
points in x-direction separated by a whitespace. The next row belong to the
following y-coordinate. Note that if the file has three or less columns, it
will be interpreted as a topography stored in a coordinate format (the three
columns contain the x, y and z coordinates of the same points). The smallest
topography that can be provided in this format is therefore 4 x 1.

The reader supports parsing file headers for additional metadata. This allows
to specify the physical size of the topography and the unit. In particular, it
supports reading ASCII files exported from SPIP and Gwyddion.

When writing your own ASCII files, we recommend to prepent the header with a
'#'. The following file is an example that contains 4 x 3 data points:
```
# Channel: Main
# Width: 10 µm
# Height: 10 µm
# Value units: m
 1.0  2.0  3.0  4.0
 5.0  6.0  7.0  8.0
 9.0 10.0 11.0 12.0
```
''')


@text
def read_xyz(fobj, physical_sizes=None, height_scale_factor=None, unit=None, info={}, periodic=None, tol=1e-6):
    """
    Load xyz-file. These files contain line scan information in terms of
    (x,y)-positions.

    Parameters
    ----------
    fobj : str or file object
         File name or stream.
    tol : float
         Tolerance for detecting uniform grids

    Returns
    -------
    topography : Topography or UniformLineScan or NonuniformLineScan
        SurfaceTopography object.
    """
    # pylint: disable=invalid-name
    data = np.loadtxt(fobj, unpack=True)

    if len(data) == 2:
        # This is a line scan.
        x, z = data
        x -= np.min(x)

        d_uniform = (x[-1] - x[0]) / (len(x) - 1)
        if np.max(np.abs(np.diff(x) - d_uniform)) < tol:
            if physical_sizes is None:
                physical_sizes = d_uniform * len(x)
            else:
                raise MetadataAlreadyFixedByFile('physical_sizes')
            t = UniformLineScan(z, physical_sizes,
                                periodic=False if periodic is None else periodic,
                                unit=unit,
                                info=info)
        else:
            if periodic is not None and periodic:
                raise ValueError('XYZ reader found nonuniform data, and the user specified that it is periodic. '
                                 'Nonuniform line scans cannot be periodic.')
            t = NonuniformLineScan(x, z, unit=unit, info=info)
            if physical_sizes is not None:
                raise MetadataAlreadyFixedByFile('physical_sizes')

    elif len(data) == 3:
        # This is a topography map.
        x, y, z = data

        # Sort values, first x than y
        indices = np.lexsort((x, y))
        x = x[indices]
        y = y[indices]
        z = z[indices]

        # Sort x-values into bins. Assume points on surface are equally spaced.
        dx = x[1] - x[0]
        binx = np.array(x / dx + 0.5, dtype=int)
        n = np.bincount(binx)
        ny = n[0]
        assert np.all(n == ny)  # FIXME: Turn assert into exception

        # Sort y-values into bins.
        dy = y[binx == 0][1] - y[binx == 0][0]
        biny = np.array(y / dy + 0.5, dtype=int)
        n = np.bincount(biny)
        nx = n[0]
        assert np.all(n == nx)  # FIXME: Turn assert into exception

        # Sort data into bins.
        data = np.zeros((nx, ny))
        data[binx, biny] = z

        # Sanity check. Should be covered by above asserts.
        value_present = np.zeros((nx, ny), dtype=bool)
        value_present[binx, biny] = True
        assert np.all(value_present)  # FIXME: Turn assert into exception

        if physical_sizes is None:
            physical_sizes = (dx * nx, dy * ny)
        else:
            raise MetadataAlreadyFixedByFile('physical_sizes')
        t = Topography(data, physical_sizes, unit=unit, info=info, periodic=periodic)
    else:
        raise Exception(
            'Expected two or three columns for topography that is a list of '
            'positions and heights.')

    if height_scale_factor is not None:
        t = t.scale(height_scale_factor)
    return t


XYZReader = make_wrapped_reader(
    read_xyz, class_name="XYZReader", format='xyz',
    name='Plain text (x,y,z coordinates)', description='''
SurfaceTopography information can be provided as coordinate data. This is a
text file that contains either two columns (for line scans) or three columns
(for two-dimensional topographies) of data. The parser does not support
reading header information. Units can therefore not be provided directly
within this file format.

Line scans can be provided on a non-uniform grid. The x-coordinates do not
need to be equally spaced and the surface specified in this format can be
reentrant. The code interprets such topographies as piecewise linear between
the points that are specified in the file.

Two-dimensional topography maps need to reside on a regular grid. The x- and
y-coordinates need to be equally spaced.
''')


def write_matrix(self, fname):
    """
    Saves the topography using `np.savetxt`. Warning: This only saves
    the heights; the physical_sizes is not contained in the file
    """
    np.savetxt(fname, self.heights())


# Register analysis functions from this module
UniformTopographyInterface.register_function('to_matrix', write_matrix)



#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Analysis of island geometries
"""

import numpy as np

from _SurfaceTopography import assign_patch_numbers as assign_patch_numbers_area  # noqa: F401
from _SurfaceTopography import assign_segment_numbers as assign_segment_numbers_area  # noqa: F401
from _SurfaceTopography import distance_map  # noqa: F401

# Stencils for determining nearest-neighbor relationships on a square grid
nn_stencil = [(1, 0), (0, 1), (-1, 0), (0, -1)]
nnn_stencil = [(1, 1), (1, 0), (1, -1), (0, -1), (-1, -1), (-1, 0), (-1, 1), (0, 1)]


def coordination(mask, periodic, stencil=nn_stencil):
    """
    Return a map with coordination numbers, i.e. the number of neighboring
    pixels where `mask` is true.

    Parameters
    ----------
    mask : array of bool
        Mask with true/false values (typically indenfying specified regions on
        a two-dimensional map)
    periodic : bool
        Mask will be treated as periodic if true.
    stencil : list of tuples
        Stencil for determining neighborhood. Each entry of the stencil
        contains relative coordinates of neighboring pixels, i.e. an
        entry (1, 0) indicates the the pixel to the right is a neighbor.

    Returns
    -------
    coordination : np.ndarray
        For each pixel, this array contains the number of neighboring pixels
        where mask is true.
    """

    coordination = np.zeros_like(mask, dtype=int)
    for dx, dy in stencil:
        tmp = np.array(mask, dtype=bool, copy=True)
        if dx != 0:
            tmp = np.roll(tmp, dx, 0)
            if not periodic:
                if dx > 0:
                    tmp[:dx, :] = 0
                else:
                    tmp[dx:, :] = 0
        if dy != 0:
            tmp = np.roll(tmp, dy, 1)
            if not periodic:
                if dy > 0:
                    tmp[:, :dy] = 0
                else:
                    tmp[:, dy:] = 0
        coordination += tmp
    return coordination


def outer_perimeter_area(c, periodic, stencil=nn_stencil):
    """
    Return a map where surface points on the outer perimeter are marked.
    """
    return np.logical_and(np.logical_not(c), coordination(c, periodic, stencil=stencil) > 0)


def inner_perimeter_area(c, periodic, stencil=nn_stencil):
    """
    Return a map where surface points on the inner perimeter are marked.
    """

    return np.logical_and(c, coordination(c, periodic, stencil=stencil) < len(stencil))


def patch_areas(patch_ids):
    """
    Return a list containing patch areas
    """
    return np.bincount(patch_ids.reshape((-1,)))[1:]


def assign_patch_numbers_profile(mask, periodic):
    patch_ids = np.cumsum(np.abs(np.diff(mask)))
    if mask[0]:
        # Patches are odd numbers
        patch_ids += 1
        if periodic and mask[-1]:
            # Assign same patch id to first and last patch
            patch_ids[patch_ids == patch_ids[-1]] = patch_ids[0]
    # Patches are odd numbers, set even numbers to zero
    patch_ids += 1
    patch_ids[(patch_ids & 0x1).astype(bool)] = 0
    patch_ids //= 2
    if mask[0]:
        patch_ids = np.append([1], patch_ids)
    else:
        patch_ids = np.append([0], patch_ids)
    nb_patches = np.max(patch_ids)
    return nb_patches, patch_ids


def outer_perimeter_profile(mask, periodic):
    p_left = np.logical_and(np.roll(mask, -1), np.logical_not(mask))
    p_right = np.logical_and(np.roll(mask, 1), np.logical_not(mask))
    if not periodic:
        p_left[-1] = False
        p_right[0] = False
    return np.logical_or(p_left, p_right)



#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Automatic determination of the reliability of the underyling data
"""

from ..HeightContainer import NonuniformLineScanInterface, UniformTopographyInterface
from ..Support.UnitConversion import get_unit_conversion_factor


def short_reliability_cutoff(self, other_cutoff=None):
    """
    Determine down to which distance scale the data is reliable, i.e. below
    which distance it may be affected by instrumental artifacts. Currently
    supported are tip radius artifacts and a user specified resolution.

    Returns
    -------
    short_cutoff : float
        All data below this length scale is unreliable. For tip radius
        analysis, this is the length of the stencil for the second derivative
        used to compute the reliability cutoff.
    """
    info = self.info
    if 'instrument' in info:
        instrument = info['instrument']
        if 'parameters' in instrument:
            parameters = instrument['parameters']
            if 'tip_radius' in parameters:
                # We are carrying out a scanning probe analysis, get tip radius in correct units
                tip_radius = parameters['tip_radius']
                r = tip_radius['value'] * get_unit_conversion_factor(tip_radius['unit'], self.unit)
                scanning_probe_cutoff = self.scanning_probe_reliability_cutoff(r)
                return scanning_probe_cutoff if other_cutoff is None else \
                    other_cutoff if scanning_probe_cutoff is None else \
                    max(scanning_probe_cutoff, other_cutoff)
            elif 'resolution' in parameters:
                # This topography just has a lateral resolution specified
                resolution = parameters['resolution']
                cutoff = resolution['value'] * get_unit_conversion_factor(resolution['unit'], self.unit)
                return cutoff if other_cutoff is None else \
                    other_cutoff if cutoff is None else \
                    max(cutoff, other_cutoff)
            else:
                # Don't know what type of instrument this is and how to carry out a reliability analysis
                return other_cutoff
        else:
            # This instrument has no parameters, we cannot carry out a reliability analysis
            return other_cutoff
    else:
        # We cannot say anything about reliability if we do not know the
        # instrument and its parameters
        return other_cutoff


UniformTopographyInterface.register_function('short_reliability_cutoff', short_reliability_cutoff)
NonuniformLineScanInterface.register_function('short_reliability_cutoff', short_reliability_cutoff)



#
# Copyright 2018-2021 Lars Pastewka
#           2019-2021 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Compute derivatives of uniform line scans and topographies"""

import numpy as np

import muFFT

from ..Exceptions import UndefinedDataError
from ..HeightContainer import UniformTopographyInterface
from ..Support import toiter
from ..UniformLineScanAndTopography import Topography

#
# Stencils for first and second derivatives
#

# First order upwind differences
first_1d = muFFT.DiscreteDerivative([0], [-1, 1])

# second order central differences of the second derivative
second_1d = muFFT.DiscreteDerivative([-1], [1, -2, 1])

# first order upwind differences of the third derivative
third_1d = muFFT.DiscreteDerivative([-1], [-1, 3, -3, 1])

# second order central differences of the third derivative
third_central_1d = muFFT.DiscreteDerivative([-2], [-1 / 2, 1, 0, -1, 1 / 2])

# First order upwind differences
first_2d_x = muFFT.DiscreteDerivative([0, 0], [[-1, 0], [1, 0]])
first_2d_y = muFFT.DiscreteDerivative([0, 0], [[-1, 1], [0, 0]])
first_2d = (first_2d_x, first_2d_y)

# second order central differences of the second derivative
second_2d_x = muFFT.DiscreteDerivative([-1, -1], [[0, 1, 0],
                                                  [0, -2, 0],
                                                  [0, 1, 0]])
second_2d_y = muFFT.DiscreteDerivative([-1, -1], [[0, 0, 0],
                                                  [1, -2, 1],
                                                  [0, 0, 0]])
second_2d = (second_2d_x, second_2d_y)

# first order upwind differences of the third derivative
third_2d_x = muFFT.DiscreteDerivative([-1, -1], [[-1],
                                                 [3],
                                                 [-3],
                                                 [1]])
third_2d_y = muFFT.DiscreteDerivative([-1, -1], [[-1, 3, -3, 1]])

third_2d = (third_2d_x, third_2d_y)


def _get_default_derivative_operator(n, dim):
    """
    Return a default set of (finite-differences) derivative operators.

    Parameters
    ----------
    n : int
        Order of derivative.
    dim : int
        Dimension of the underlying field. (For dim > 1, this returns a
        representation of the nabla-operator.)
    """
    if dim == 1:
        if n == 1:
            return first_1d
        elif n == 2:
            return second_1d
        elif n == 3:
            return third_1d
        else:
            raise ValueError("Don't know how to compute derivative of order "
                             "{}.".format(n))
    elif dim == 2:
        if n == 1:
            return first_2d
        elif n == 2:
            return second_2d
        elif n == 3:
            return third_2d
        else:
            raise ValueError("Don't know how to compute derivative of order "
                             "{}.".format(n))
    else:
        raise ValueError("Don't know how to compute the derivative of a "
                         "topography of dimension {}.".format(dim))


def trim_nonperiodic(arr, scale_factor, op):
    """
    Trim the outer edges of an array that contains derivatives computed under
    the assumption of periodicity. (These values at the outer edge will simply
    be wrong.)

    Parameters
    ----------
    arr : np.ndarray
        Array to be trimmed.
    scale_factor : int, optional
        Integer factor that scales the stencil difference.
    op : muFFT.DiscreteDerivative
        Derivative operator that contains information about the size of the
        stencil.
    """
    if not isinstance(op, muFFT.DiscreteDerivative):
        raise ValueError('Can only trim edges for discrete derivatives.')

    lbounds = np.array(op.lbounds)
    rbounds = lbounds + np.array(op.stencil.shape)

    # Loop over dimension and add slicing information to `trimmed_slice`
    trimmed_slice = []
    for left, right, s in zip(lbounds, rbounds, scale_factor):
        # This is the leftmost distance in the stencil from the point where the derivative is computed
        # This value is always positive (or zero), i.e. it truncates the start of the array
        left = np.ceil(-s * min(left, 0))
        # This is the rightmost distance in the stencil from the point where the derivative is computed
        # This value is always negative (or zero), i.e. it truncates the end of the array
        right = np.floor(-s * max(0, right - 1))
        # If right is zero, we set it to None to indicate that nothing is truncated from the end of the array
        if right == 0:
            right = None
        trimmed_slice += [slice(int(left), None if right is None else int(right))]

    return arr[tuple(trimmed_slice)]


def derivative(self, n, scale_factor=None, distance=None, operator=None, periodic=None, mask_function=None,
               interpolation='linear', progress_callback=None):
    """
    Compute derivative of topography or line scan stored on a uniform grid.

    Parameters
    ----------
    self : :obj:`SurfaceTopography` or :obj:`UniformLineScan`
        Surface topography object containing height information.
    n : int
        Order of the derivative.
    scale_factor : int or list of ints or list of tuples of ints, optional
        Integer factor that scales the stencil difference, i.e.
        specifying 2 will compute the derivative using a discrete step of
        2 * px. Either `scale_factor` or `distance` can be specified.
            - Single int: Returns a single derivative scaled in all directions
              with this value
            - List of ints: Returns multiple derivatives, each scaled in all
              direction with the respective value from the list
            - List of tuples of ints: Each tuple contains a scale factor in
              the two Cartesian (x- and y-) directions. Return multiple
              derivatives, scaled with different factors in both directions.
        (Default: None)
    distance : float or list of floats, optional
        Explicit distance scale for computation of the derivative. Either
        `scale_factor` or `distance` can be specified. Note that the distance
        specifies the overall length of the stencil of lowest truncation
        order, not the effective grid spacing used by this stencil. The scale
        factor is then given by distance / (n * px) where n is the order of the
        derivative and px the grid spacing.
        (Default: None)
    operator : :obj:`muFFT.Derivative` object or tuple of :obj:`muFFT.Derivative` objects, optional
        Derivative operator used to compute the derivative. If unspecified,
        a simple upwind differences scheme will be applied to compute the
        derivative. A tuple contains the gradient, i.e. the derivative
        operators in the Cartesian directions for multidimensional fields.
        (Default: None)
    periodic : bool, optional
        Override periodic flag from topography. (Default: None)
    mask_function : function, optional
        A function that takes as argument the output of FFT.fftfreq and
        returns a mask that will be multiplied with the Fourier transformed
        topography. This can be used to implement Fourier filtering before
        computing the derivative. (Default: None)
    interpolation : str, optional
        Interpolation method to use for fractional scale factors. Use
        'linear' for a local liner interpolation or 'fourier' for global
        Fourier interpolation. Set to 'disable' to raise an error when
        interpolation is necessary. Note that Fourier interpolation carries
        large errors for nonperiodic topographies and should be used with
        care. (Default: 'linear')
    progress_callback : func, optional
        Function taking iteration and the total number of iterations as
        arguments for progress reporting. (Default: None)

    Returns
    -------
    derivative : array or tuple of arrays
        Array with derivative values. If dimension of the topography is
        unity (line scan), then an array of the same shape as the
        topography is returned. Otherwise, the first array index contains
        the direction of the derivative. If the topgography is nonperiodic,
        then all returning array with have shape one less than the input
        arrays.
    """
    if self.physical_sizes is None:
        raise ValueError(
            'SurfaceTopography does not have physical size information, but '
            'this is required to be able to compute a derivative.')

    if self.has_undefined_data:
        raise UndefinedDataError('This topography has undefined data (missing data points). Derivatives cannot be '
                                 'computed for topographies with missing data points.')

    if operator is None:
        operator = _get_default_derivative_operator(n, self.dim)

    if interpolation == 'linear':
        linear = self.interpolate_linear()
        positions = np.transpose(self.positions())
    elif interpolation == 'fourier' or interpolation == 'disable':
        linear = None
    else:
        raise ValueError("`interpolation` argument must be either 'linear', 'fourier' or 'disable'.")

    pixel_size = np.array(self.pixel_size)

    nb_scale_factors = 1
    scale_factor_is_array = False
    if scale_factor is None:
        if distance is None:
            # This is the default behavior
            scale_factor = [1]
        else:
            # Convert distance to scale factor
            if self.dim == 1:
                px, = pixel_size
                try:
                    scale_factor = [d / (n * px) for d in distance]
                    nb_scale_factors = len(scale_factor)
                    scale_factor_is_array = True
                except TypeError:
                    scale_factor = [distance / (n * px)]
            else:
                px, py = pixel_size
                try:
                    scale_factor = [(d / (n * px), d / (n * py)) for d in distance]
                    nb_scale_factors = len(scale_factor)
                    scale_factor_is_array = True
                except TypeError:
                    scale_factor = [(distance / (n * px), distance / (n * py))]
    elif distance is not None:
        raise ValueError('Please specify either `scale_factor` or `distance`')
    else:
        try:
            iter(scale_factor)
            nb_scale_factors = len(scale_factor)
            scale_factor_is_array = True
        except TypeError:
            scale_factor = [scale_factor]

    is_periodic = self.is_periodic if periodic is None else periodic

    # Return FFT object (this will only be initialized once and reused in subsequent calls)
    fft = self.make_fft()

    # These fields are reused when this function is called multiple times
    real_field = fft.fetch_or_register_real_space_field('real_temporary', 1)
    fourier_field = fft.fetch_or_register_fourier_space_field('complex_temporary', 1)
    np.array(real_field, copy=False)[...] = self.heights()
    fft.fft(real_field, fourier_field)

    # Apply mask function
    if mask_function is not None:
        np.array(fourier_field, copy=False)[...] *= mask_function((fft.fftfreq.T / pixel_size).T)

    fourier_array = np.array(fourier_field, copy=False)
    fourier_copy = fourier_array.copy()

    # Apply derivative operator in Fourier space
    derivatives = []
    operators = toiter(operator)
    for i, op in enumerate(operators):
        der = []
        for j, s in enumerate(scale_factor):
            if progress_callback is not None:
                progress_callback(i * nb_scale_factors + j, len(operators) * nb_scale_factors)
            s = np.array(s) * np.ones_like(pixel_size)
            scaled_pixel_size = s * pixel_size
            interpolation_required = np.any(s - s.astype(int) != 0)
            if interpolation_required and interpolation == 'disabled':
                raise ValueError('Interpolation is required to compute derivative at the desired scale but is '
                                 'explicitly disabled through the `interpolation` argument.')
            if interpolation_required and linear is not None:
                # We need to interpolate using the linear interpolator
                lbounds = np.array(op.lbounds)
                stencil = np.array(op.stencil)
                _der = np.zeros_like(real_field)
                for stencil_coordinate, stencil_value in np.ndenumerate(stencil):
                    if stencil_value:
                        stencil_positions = positions + (lbounds + stencil_coordinate) * scaled_pixel_size
                        # We enforce periodicity here but will trim the (erroneous) boundary region below
                        if self.dim == 1:
                            _der += stencil_value * linear(stencil_positions, periodic=True)
                        else:
                            _der += stencil_value * linear(*stencil_positions.T, periodic=True)

            else:
                # We can use the Fourier trick to compute the derivative; this gives the exact stencil derivative if
                # interpolation is not required and the Fourier-interpolated derivative for fractional scale factors
                fourier_array[...] = fourier_copy * op.fourier((fft.fftfreq.T * s).T)
                fft.ifft(fourier_field, real_field)
                _der = np.array(real_field, copy=False) * fft.normalisation

            if not is_periodic:
                _der = trim_nonperiodic(_der, s, op)

            # We need to divide by the grid spacing to make this a derivative
            _der /= scaled_pixel_size[i] ** n

            if scale_factor_is_array:
                der += [_der]
            else:
                der = _der
        try:
            iter(operator)
            derivatives += [der]
        except TypeError:
            derivatives = der
    if progress_callback is not None:
        progress_callback(len(operators) * nb_scale_factors, len(operators) * nb_scale_factors)
    return derivatives


def fourier_derivative(self, scale_factor=None, distance=None, mask_function=None):
    r"""
    First derivatives of the fourier interpolation of the topography.

    For example in x direction:

    .. math::

        \left[ \frac{\partial h }{\partial x}\right]_{x_i, y_j} =
        \frac{1}{2 \pi }
        \sum_{kl} q_x^k \tilde h_{kl} e^{i (q_x^k x_i + q_y^l y_j}


    With :math:`\tilde h_{kl}` the DFT of the heights :math:``h
    For even number of points, the interpretation of the components at the
    Niquist frequency is ambiguous (we are free to choose amplitude or phase)
    Following (https://math.mit.edu/~stevenj/fft-deriv.pdf),
    we assume it is cosinusoidal.

    Parameters
    ----------
    self : :obj:`SurfaceTopography` or :obj:`UniformLineScan`
        Surface topography object containing height information.
    scale_factor : int or list of ints or list of tuples of ints, optional
        Integer factor that scales the stencil difference, i.e.
        specifying 2 will compute the derivative using a discrete step of
        2 * px. Either `scale_factor` or `distance` can be specified.
            - Single int: Returns a single derivative scaled in all directions
              with this value
            - List of ints: Returns multiple derivatives, each scaled in all
              direction with the respective value from the list
            - List of tuples of ints: Each tuple contains a scale factor in
              the two Cartesian (x- and y-) directions. Return multiple
              derivatives, scaled with different factors in both directions.
        (Default: None)
    distance : float or list of floats, optional
        Explicit distance scale for computation of the derivative. Either
        `scale_factor` or `distance` can be specified. Note that the distance
        specifies the overall length of the stencil of lowest truncation
        order, not the effective grid spacing used by this stencil. The scale
        factor is then given by distance / (n * px) where n is the order of the
        derivative and px the grid spacing.
        (Default: None)
    mask_function : function, optional
        A function that takes as argument the output of FFT.fftfreq and
        returns a mask that will be multiplied with the Fourier transformed
        topography. This can be used to implement Fourier filtering before
        computing the derivative. (Default: None)

    Returns
    -------
    derivative : array or tuple of arrays
        Array with derivative values. If dimension of the topography is
        unity (line scan), then an array of the same shape as the
        topography is returned. Otherwise, the first array index contains
        the direction of the derivative. If the topgography is nonperiodic,
        then all returning array with have shape one less than the input
        arrays.
    """
    dim = self.dim
    return self.derivative(1, operator=(muFFT.FourierDerivative(dim, i) for i in range(dim)),
                           scale_factor=scale_factor, distance=distance, mask_function=mask_function)


# Register analysis functions from this module
UniformTopographyInterface.register_function('derivative', derivative)
Topography.register_function('fourier_derivative', fourier_derivative)



#
# Copyright 2020-2021 Lars Pastewka
#           2019-2020 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from .Bibliography import doi  # noqa: F401
from .Deprecation import deprecated, DeprecatedDictionary  # noqa: F401
from .Interpolation import Bicubic  # noqa: F401
from .Regression import resample_radial  # noqa: F401


def toiter(obj):
    """If `obj` is scalar, wrap it into a list"""
    try:
        iter(obj)
        return obj
    except TypeError:
        return [obj]


def fromiter(result, obj):
    """If `obj` is scalar, return first element of result list"""
    try:
        iter(obj)
        return result
    except TypeError:
        return result[0]


def build_tuple(quantities, **kwargs):
    """
    Build a custom tuple, where each of the characters in quantities
    represents a certain datum.

    Parameters
    ----------
    quantities : str
        String with quantities, e.g. 'dhj' returns a 3-tuple with the
        datums associated with the characters 'd', 'h' and 'j'.
    **kwargs : dict
        Individual datums, e.g. `d=..., h=..., j=...`.

    Returns
    -------
    datum : tuple
        Tuple of length `quantities`.
    """
    retvals = []
    for c in quantities:
        retvals += [kwargs[c]]
    return tuple(retvals)



#
# Copyright 2019-2021 Lars Pastewka
#           2019-2021 Michael Röttger
#           2019 Kai Haase
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import numpy as np

from igor.binarywave import load as loadibw

from ..Exceptions import MetadataAlreadyFixedByFile
from ..UniformLineScanAndTopography import Topography

from .common import OpenFromAny
from .Reader import ReaderBase, ChannelInfo


class IBWReader(ReaderBase):
    _format = 'ibw'
    _name = 'Igor binary wave'
    _description = '''
Igor binary wave is a container format of the
[Igor Pro](https://www.wavemetrics.com/products/igorpro/programming)
language. This format is used by AFMs from Asylum Research (now Oxford
Instruments) to store topography information. This format contains information
on the physical size of the topography map as well as its units.
'''

    # Reads in the positions of all the data and metadata
    def __init__(self, file_path):
        with OpenFromAny(file_path, 'rb') as f:
            file = loadibw(f)

        if file['version'] != 5:
            raise RuntimeError('Only IBW version 5 is supported!')

        self.data = file['wave']

        # the first two labels are of x and y axis. we cannot read those
        self._channel_names = [channel.decode() for channel in
                               self.data['labels'][2][1:]]
        self._default_channel = 0

        #
        # Read physical sizes from file, since all data
        # has already been loaded, we can calculate it here
        #
        height_data = self.data['wData']

        # TODO is it always like this?
        assert len(height_data.shape) == 3, \
            "We expect all channels being coded in to one wave and all are " \
            "2D. This is not true somehow, cannot proceed."
        nx, ny, num_channels = height_data.shape

        # ensure that there are not too many channel names
        self._channel_names = self._channel_names[:num_channels]

        # add channel names for all channels without name
        no_name_idx = 1
        while len(self._channel_names) < num_channels:
            self._channel_names.append("no name ({})".format(no_name_idx))
            no_name_idx += 1

        #
        # Decode units
        #
        def decode_unit_entry(u):
            return u.tobytes().partition(b'\0')[0].decode('latin-1')

        data_unit = decode_unit_entry(self.data['wave_header']['dataUnits'])
        x_unit = decode_unit_entry(self.data['wave_header']['dimUnits'][0])
        y_unit = decode_unit_entry(self.data['wave_header']['dimUnits'][1])

        # the following is not necessary, we could handle different units by
        # rescaling, however I'll leave it like this for now, print some
        # message so we know what to do if a file occurs which does not fulfill
        # this assumption
        assert data_unit == x_unit == y_unit, \
            "So far, data units and dimension units must be all the same. " +\
            "data unit: '{}', x unit: '{}', y unit: '{}'".format(data_unit,
                                                                 x_unit,
                                                                 y_unit)

        # An empty unit should be None
        if data_unit == '':
            data_unit = None

        self._data_unit = data_unit

        #
        # Decode sizes
        #
        sfA = self.data['wave_header']['sfA']
        self._physical_sizes = (nx * sfA[0], ny * sfA[1])
        # Comment in C header file on these fields: Index value for element e
        # of dimension d = sfA[d]*e + sfB[d]. sfB is left out here, because we
        # are interested in the width and height, not the absolute offsets.

        #
        # Build channel information
        #
        self._channels = [
            ChannelInfo(self, i, name=cn, dim=2, nb_grid_pts=(nx, ny), physical_sizes=self._physical_sizes,
                        unit=self._data_unit, height_scale_factor=1, uniform=True)
            for i, cn in enumerate(self._channel_names)]

        # Shall we use the channel names in order to assign a unit as Gwyddion
        # does?

    @property
    def channels(self):
        return self._channels

    def topography(self, channel_index=None, physical_sizes=None,
                   height_scale_factor=None, unit=None, info={},
                   periodic=False, subdomain_locations=None,
                   nb_subdomain_grid_pts=None):
        if channel_index is None:
            channel_index = self._default_channel_index

        if unit is not None and self._data_unit is not None:
            raise MetadataAlreadyFixedByFile('unit')
        if self._data_unit is not None:
            unit = self._data_unit

        if height_scale_factor is not None:
            raise MetadataAlreadyFixedByFile('height_scale_factor')

        if subdomain_locations is not None or \
                nb_subdomain_grid_pts is not None:
            raise RuntimeError(
                'This reader does not support MPI parallelization.')

        height_data = self.data['wData']
        height_data = np.fliplr(height_data[:, :, channel_index].copy())

        if physical_sizes is None:
            physical_sizes = self._physical_sizes
        else:
            raise MetadataAlreadyFixedByFile('physical_sizes')

        topo = Topography(height_data, physical_sizes, unit=unit, info=info, periodic=periodic)
        # we could pass the data units here, but they dont seem to be always
        # correct for all channels?!

        channel = self._channels[channel_index]
        if channel.height_scale_factor is not None:
            if height_scale_factor is not None:
                raise MetadataAlreadyFixedByFile('height_scale_factor')
            height_scale_factor = channel.height_scale_factor

        if height_scale_factor is not None:
            topo = topo.scale(height_scale_factor)

        return topo



#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

#
# Deep Zoom Image File Format (DZI) - The DZI file format is described in detail here:
# https://github.com/openseadragon/openseadragon/wiki/The-DZI-File-Format
# https://docs.microsoft.com/en-us/previous-versions/windows/silverlight/dotnet-windows-silverlight/cc645077(v=vs.95)
#

import json
import math
import os
import xml.etree.cElementTree as ET

import numpy as np
from matplotlib import cm
from PIL import Image

from numpyencoder import NumpyEncoder

from ..HeightContainer import UniformTopographyInterface
from ..Support.UnitConversion import get_unit_conversion_factor, suggest_length_unit_for_data


def write_dzi(data, name, physical_sizes, unit, root_directory='.', tile_size=256, overlap=1, format='jpg',
              meta_format='xml', colorbar_title=None, cmap=None, **kwargs):
    """
    Write generic numpy array to a Deep Zoom Image file. This can for example
    be used to create a zoomable topography with OpenSeadragon
    (https://openseadragon.github.io/).

    Additional keyword parameters are passed to Pillow's `save` function.

    Parameters
    ----------
    data : np.ndarray
        Two-dimensional array containing the data.
    name : str
        Name of the exported file. This is used as a prefix. Output filter
        create the file `name`.xml that contains the metadata and a directory
        `name`_files that contains the rendered image files at different levels.
    physical_sizes : tuple of floats
        Linear physical sizes of the two-dimensional array.
    unit : str
        Length units of physical sizes.
    root_directory : str, optional
        Root directory where to place `name`.xml and `name`_files.
        (Default: '.')
    tile_size : int, optional
        Size of individual tiles. (Default: 256)
    overlap : int, optional
        Overlap of tiles. (Default: 1)
    format : str, optional
        Image format. Note that PNG files have seams at the boundary between
        tiles. (Default: 'jpg')
    meta_format : str, optional
        Format for metadata information (the DZI file), can be 'xml' or
        'json'. (Default: 'xml')
    colorbar_title : str, optional
        Additional title for the color bar that is dumped into the DZI file.
        (Default: None)
    cmap : str or colormap, optional
        Color map for rendering the topography. (Default: None)

    Returns
    -------
    manifest : list of str
        List with names of files created during write operation
    """
    cmap = cm.get_cmap(cmap)

    # Image size
    full_width, full_height = width, height = data.shape

    # Compute pixels per meter
    sx, sy = physical_sizes
    fac = get_unit_conversion_factor(unit, 'm')
    pixels_per_meter_width = width / (fac * sx)
    pixels_per_meter_height = height / (fac * sy)

    # Get heights and rescale to interval 0, 1
    mx, mn = data.max(), data.min()
    data = (data - mn) / (mx - mn)

    # Write configuration file
    if meta_format == 'xml':
        fn = os.path.join(root_directory, name + '.xml')
        root = ET.Element('Image', TileSize=str(tile_size), Overlap=str(overlap), Format=format, Colormap=cmap.name,
                          xmlns='http://schemas.microsoft.com/deepzoom/2008')
        if colorbar_title is not None:
            root.set('ColorbarTitle', colorbar_title)
        ET.SubElement(root, 'Size', Width=str(width), Height=str(height))
        ET.SubElement(root, 'PixelsPerMeter', Width=str(pixels_per_meter_width), Height=str(pixels_per_meter_height))
        ET.SubElement(root, 'ColorbarRange', Minimum=str(mn), Maximum=str(mx))
        os.makedirs(root_directory, exist_ok=True)
        ET.ElementTree(root).write(fn, encoding='utf-8', xml_declaration=True)
    elif meta_format == 'json':
        fn = os.path.join(root_directory, name + '.json')
        with open(fn, 'w') as f:
            image_dict = {
                'xmlns': 'http://schemas.microsoft.com/deepzoom/2008',
                'Format': format,
                'Overlap': overlap,
                'TileSize': tile_size,
                'Size': {
                    'Width': width,
                    'Height': height
                },
                'PixelsPerMeter': {
                    'Width': pixels_per_meter_width,
                    'Height': pixels_per_meter_height
                },
                'Colormap': cmap.name,
                'ColorbarRange': {
                    'Minimum': mn,
                    'Maximum': mx
                }
            }
            if colorbar_title is not None:
                image_dict.update({'ColorbarTitle': colorbar_title})
            json.dump({'Image': image_dict}, f, cls=NumpyEncoder)
    else:
        raise ValueError(f'Unknown metadata format {meta_format}.')
    manifest = [fn]

    # Determine number of levels
    max_level = math.ceil(math.log2(max(width, height)))

    # Loop over levels and write tiles
    root_directory = os.path.join(root_directory, name + '_files')
    os.makedirs(root_directory, exist_ok=True)
    step = 1
    for level in range(max_level, -1, -1):
        level_root_directory = os.path.join(root_directory, str(level))
        os.makedirs(level_root_directory, exist_ok=True)

        columns = math.ceil(width / tile_size)
        rows = math.ceil(height / tile_size)

        # Loop over all tiles
        for column in range(columns):
            for row in range(rows):
                # File name for this tile
                fn = os.path.join(level_root_directory, f'{column}_{row}.{format}')

                # Determine image section of this tile
                left = (column * tile_size - overlap) * step
                bottom = (row * tile_size - overlap) * step

                right = ((column + 1) * tile_size + overlap) * step
                top = ((row + 1) * tile_size + overlap) * step

                if left < 0:
                    left = 0
                if bottom < 0:
                    bottom = 0
                if right > full_width - 1:
                    right = full_width - 1
                if top > full_height - 1:
                    top = full_height - 1

                # Convert to image and save
                colors = (cmap(data[left:right:step, bottom:top:step].T) * 255).astype(np.uint8)
                # Remove alpha channel before writing
                Image.fromarray(colors[:, :, :3]).save(fn, **kwargs)
                manifest += [fn]

        width = math.ceil(width / 2)
        height = math.ceil(height / 2)
        step *= 2

    return manifest


def write_topography_dzi(self, name, root_directory='.', tile_size=256, overlap=1, format='jpg', meta_format='xml',
                         cmap=None, **kwargs):
    """
    Write topography to a Deep Zoom Image file. This can for example be used
    to create a zoomable topography with OpenSeadragon
    (https://openseadragon.github.io/).

    Additional keyword parameters are passed to Pillow's `save` function.

    Parameters
    ----------
    self : :obj:`Topography`
        Topogaphy to export
    name : str
        Name of the exported file. This is used as a prefix. Output filter
        create the file `name`.xml that contains the metadata and a directory
        `name`_files that contains the rendered image files at different levels.
    root_directory : str
        Root directory where to place `name`.xml and `name`_files.
    tile_size : int, optional
        Size of individual tiles. (Default: 256)
    overlap : int, optional
        Overlap of tiles. (Default: 1)
    format : str, optional
        Image format. Note that PNG files have seems at the boundary between
        tiles. (Default: jpg)
    meta_format : str, optional
        Format for metadata information (the DZI file), can be 'xml' or
        'json'. (Default: 'xml')
    cmap : str or colormap, optional
        Color map for rendering the topography. (Default: None)

    Returns
    -------
    filenames : list of str
        List with names of files created during write operation
    """
    # Get reasonable unit
    ideal_height_unit = suggest_length_unit_for_data('linear', self.heights(), self.unit)
    t = self.to_unit(ideal_height_unit)
    return write_dzi(t.heights(), name, t.physical_sizes, ideal_height_unit, root_directory=root_directory,
                     tile_size=tile_size, overlap=overlap, format=format, meta_format=meta_format,
                     colorbar_title=f'Height ({ideal_height_unit})', cmap=cmap, **kwargs)


UniformTopographyInterface.register_function('to_dzi', write_topography_dzi)



#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
This module contains most exceptions used in SurfaceTopography.
"""


class CannotPerformAnalysisError(Exception):
    """
    Exception raised when an analysis cannot be performed.
    """
    pass


class NoReliableDataError(CannotPerformAnalysisError):
    """
    Exception indicates that an analysis has no reliable data and would need
    to return an empty result.
    """
    pass


class ReentrantDataError(CannotPerformAnalysisError):
    """
    Exception indicates that the underlying data is reentrant and that the
    analysis function does not work on reentrant data.
    """
    pass


class UndefinedDataError(CannotPerformAnalysisError):
    """
    Exception indicates that the underlying data has missing (undefined) data
    points and that the analysis function cannot handle missing data
    """
    pass


# I/O exceptions

class ReadFileError(Exception):
    pass


class UnknownFileFormatGiven(ReadFileError):
    pass


class CannotDetectFileFormat(ReadFileError):
    """
    Raised when no reader is able to open_topography the file
    """
    pass


class FileFormatMismatch(ReadFileError):
    """
    Raised when the reader cannot interpret the file at all
    (obvious for txt vs binary, but holds also for a header)
    """
    pass


class CorruptFile(ReadFileError):
    """
    Raised when the reader identifies the file format as matching,
    but there is a mistake, for example the number of points doesn't match
    """
    pass


class MetadataAlreadyFixedByFile(ReadFileError):
    """
    Raised when instantiating a topography from a reader,
    given metadata which cannot be overridden, because
    it is already fixed by the file contents.
    """

    def __init__(self, kw, alt_msg=None):
        """
        Parameters
        ----------
        kw: str
            Name of the keyword argument to .topography().
        alt_msg: str or None
            If not None, use this as error message instead of
            the default one.
        """
        self._kw = kw
        self._alt_msg = alt_msg

    def __str__(self):
        if self._alt_msg:
            return self._alt_msg
        else:
            return f"Value for keyword '{self._kw}' is already fixed by file contents and cannot be overridden"



#
# Copyright 2019, 2021 Michael Röttger
#           2015-2016, 2018-2021 Lars Pastewka
#           2018-2019 Antoine Sanner
#           2015-2016 Till Junge
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Functions computing scalar roughness parameters
"""

import numpy as np

from NuMPI.Tools import Reduction

from ..HeightContainer import UniformTopographyInterface


def rms_height_from_profile(topography):
    """
    Compute the root mean square height amplitude of a topography or
    line scan stored on a uniform grid from individual profiles.
    (This is the Rq value.)

    Parameters
    ----------
    topography : :obj:`SurfaceTopography` or :obj:`UniformLineScan`
        SurfaceTopography object containing height information.

    Returns
    -------
    rms_height : float
        Root mean square height value.
    """
    if topography.is_domain_decomposed:
        raise NotImplementedError('`rms_height_from_profile` does not support MPI-decomposed topographies.')

    n = np.prod(topography.nb_grid_pts)
    reduction = Reduction(topography._communicator)
    profile = topography.heights()
    return np.sqrt(reduction.sum((profile - reduction.sum(profile, axis=0) / topography.nb_grid_pts[0]) ** 2) / n)


def rms_height_from_area(topography):
    """
    Compute the root mean square height amplitude of a topography or
    line scan stored on a uniform grid from the whole areal data.
    (This is the Sq value.)

    Parameters
    ----------
    topography : :obj:`SurfaceTopography` or :obj:`UniformLineScan`
        SurfaceTopography object containing height information.

    Returns
    -------
    rms_height : float
        Root mean square height value.
    """
    if topography.dim <= 1:
        raise ValueError('Areal rms height can only be computed for topographies, not line scans.')
    elif topography.dim == 2:
        n = np.prod(topography.nb_grid_pts)
        reduction = Reduction(topography._communicator)
        profile = topography.heights()
        return np.sqrt(reduction.sum((profile - reduction.sum(profile) / n) ** 2) / n)
    else:
        raise ValueError(f'Cannot handle topographies of dimension {topography.dim}')


def rms_gradient(topography, short_wavelength_cutoff=None, window=None,
                 direction=None):
    """
    Compute the root mean square amplitude of the height gradient of a
    topography stored on a uniform grid. The topography must be
    two-dimensional (i.e. a topography map).

    Parameters
    ----------
    topography : :obj:`SurfaceTopography` or :obj:`UniformLineScan`
        SurfaceTopography object containing height information.
    short_wavelength_cutoff : float
        All wavelengths below this cutoff will be set to zero amplitude.
    window : str, optional
        Window for eliminating edge effect. See scipy.signal.get_window.
        Only used if short wavelength cutoff is set.
        (Default: no window for periodic Topographies, "hann" window for
        nonperiodic Topographies)
    direction : str, optional
        Direction in which the window is applied. Possible options are
        'x', 'y' and 'radial'. If set to None, it chooses 'x' for line
        scans and 'radial' for topographies. Only used if short wavelength
        cutoff is set. (Default: None)

    Returns
    -------
    rms_slope : float
        Root mean square slope value.
    """
    if short_wavelength_cutoff is not None:
        topography = topography.window(window=window, direction=direction)
    if topography.dim <= 1:
        raise ValueError('RMS gradient can only be computed for topographies, not line scans.')
    elif topography.dim == 2:
        mask_function = None if short_wavelength_cutoff is None else \
            lambda frequency: (frequency[0] ** 2 + frequency[1] ** 2) < 1 / short_wavelength_cutoff ** 2
        slx, sly = topography.derivative(1, mask_function=mask_function)
        reduction = Reduction(topography._communicator)
        return np.sqrt(reduction.mean(slx ** 2 + sly ** 2))
    else:
        raise ValueError(f'Cannot handle topographies of dimension {topography.dim}')


def rms_slope_from_profile(topography, short_wavelength_cutoff=None, window=None,
                           direction=None):
    """
    Compute the root mean square amplitude of the height derivative of a
    topography or line scan stored on a uniform grid. If the topography is two
    dimensional (i.e. a topography map), the derivative is computed in the
    x-direction.

    Parameters
    ----------
    topography : :obj:`SurfaceTopography` or :obj:`UniformLineScan`
        SurfaceTopography object containing height information.
    short_wavelength_cutoff : float
        All wavelengths below this cutoff will be set to zero amplitude.
    window : str, optional
        Window for eliminating edge effect. See scipy.signal.get_window.
        Only used if short wavelength cutoff is set.
        (Default: no window for periodic Topographies, "hann" window for
        nonperiodic Topographies)
    direction : str, optional
        Direction in which the window is applied. Possible options are
        'x', 'y' and 'radial'. If set to None, it chooses 'x' for line
        scans and 'radial' for topographies. Only used if short wavelength
        cutoff is set. (Default: None)

    Returns
    -------
    rms_slope : float
        Root mean square slope value.
    """
    if short_wavelength_cutoff is not None:
        topography = topography.window(window=window, direction=direction)
    mask_function = None if short_wavelength_cutoff is None else \
        lambda frequency: frequency[0] ** 2 < 1 / short_wavelength_cutoff ** 2
    if topography.dim == 1:
        dx = topography.derivative(1, mask_function=mask_function)
    elif topography.dim == 2:
        dx, dy = topography.derivative(1, mask_function=mask_function)
    else:
        raise ValueError(f'Cannot handle topographies of dimension {topography.dim}')
    reduction = Reduction(topography._communicator)
    return np.sqrt(reduction.mean(dx ** 2))


def rms_curvature_from_profile(topography, short_wavelength_cutoff=None, window=None,
                               direction=None):
    """
    Compute the root mean square amplitude of the second derivative (i.e. the
    curvature) of a topography or line scan stored on a uniform grid. If the
    topography is two-dimensional (i.e. a topography map), then the rms curvature
    is computed only along the x-direction.

    Parameters
    ----------
    topography : :obj:`SurfaceTopography` or :obj:`UniformLineScan`
        SurfaceTopography object containing height information.
    short_wavelength_cutoff : float
        All wavelengths below this cutoff will be set to zero amplitude.
    window : str, optional
        Window for eliminating edge effect. See scipy.signal.get_window.
        Only used if short wavelength cutoff is set.
        (Default: no window for periodic Topographies, "hann" window for
        nonperiodic Topographies)
    direction : str, optional
        Direction in which the window is applied. Possible options are
        'x', 'y' and 'radial'. If set to None, it chooses 'x' for line
        scans and 'radial' for topographies. Only used if short wavelength
        cutoff is set. (Default: None)

    Returns
    -------
    rms_curvature : float
        Root mean square curvature value.
    """
    if short_wavelength_cutoff is not None:
        topography = topography.window(window=window, direction=direction)
    mask_function = None if short_wavelength_cutoff is None else \
        lambda frequency: frequency[0] ** 2 < 1 / short_wavelength_cutoff ** 2
    if topography.dim == 1:
        d2x = topography.derivative(2, mask_function=mask_function)
    elif topography.dim == 2:
        d2x, d2y = topography.derivative(2, mask_function=mask_function)
    else:
        raise ValueError(f'Cannot handle topographies of dimension {topography.dim}')
    reduction = Reduction(topography._communicator)
    return np.sqrt(reduction.mean(d2x ** 2))


def rms_laplacian(topography, short_wavelength_cutoff=None, window=None,
                  direction=None):
    """
    Compute the root mean square amplitude of the Laplacian (i.e. the sum of
    second derivatives in x- and y-directions) of a topography on a uniform
    grid. The topography must be two-dimensional (i.e. a topography map).

    Parameters
    ----------
    topography : :obj:`SurfaceTopography` or :obj:`UniformLineScan`
        SurfaceTopography object containing height information.
    short_wavelength_cutoff : float
        All wavelengths below this cutoff will be set to zero amplitude.
    window : str, optional
        Window for eliminating edge effect. See scipy.signal.get_window.
        Only used if short wavelength cutoff is set.
        (Default: no window for periodic Topographies, "hann" window for
        nonperiodic Topographies)
    direction : str, optional
        Direction in which the window is applied. Possible options are
        'x', 'y' and 'radial'. If set to None, it chooses 'x' for line
        scans and 'radial' for topographies. Only used if short wavelength
        cutoff is set. (Default: None)

    Returns
    -------
    rms_laplacian : float
        Root mean square Laplacian value.
    """
    if short_wavelength_cutoff is not None:
        topography = topography.window(window=window, direction=direction)
    if topography.dim == 1:
        raise ValueError('RMS Laplacian can only be computed for topographies, not line scans.')
    elif topography.dim == 2:
        mask_function = None if short_wavelength_cutoff is None else \
            lambda frequency: (frequency[0] ** 2 + frequency[1] ** 2) < 1 / short_wavelength_cutoff ** 2
        curv = topography.derivative(2, mask_function=mask_function)
        reduction = Reduction(topography._communicator)
        return np.sqrt(reduction.mean((curv[0] + curv[1]) ** 2))
    else:
        raise ValueError(f'Cannot handle topographies of dimension {topography.dim}')


def rms_curvature_from_area(topography, short_wavelength_cutoff=None, window=None,
                            direction=None):
    """
    Compute the root mean square amplitude of the curvature of a
    topography stored on a uniform grid. The topography must be
    two-dimensional (i.e. a topography map). This function returns
    half the Laplacian.

    Parameters
    ----------
    topography : :obj:`SurfaceTopography` or :obj:`UniformLineScan`
        SurfaceTopography object containing height information.
    short_wavelength_cutoff : float
        All wavelengths below this cutoff will be set to zero amplitude.
    window : str, optional
        Window for eliminating edge effect. See scipy.signal.get_window.
        Only used if short wavelength cutoff is set.
        (Default: no window for periodic Topographies, "hann" window for
        nonperiodic Topographies)
    direction : str, optional
        Direction in which the window is applied. Possible options are
        'x', 'y' and 'radial'. If set to None, it chooses 'x' for line
        scans and 'radial' for topographies. Only used if short wavelength
        cutoff is set. (Default: None)

    Returns
    -------
    rms_curvature : float
        Root mean square curvature value.
    """
    return rms_laplacian(topography, short_wavelength_cutoff=short_wavelength_cutoff, window=window,
                         direction=direction) / 2


# Register analysis functions from this module
UniformTopographyInterface.register_function('rms_height_from_profile', rms_height_from_profile)
UniformTopographyInterface.register_function('rms_height_from_area', rms_height_from_area)
UniformTopographyInterface.register_function('rms_gradient', rms_gradient)
UniformTopographyInterface.register_function('rms_slope_from_profile', rms_slope_from_profile)
UniformTopographyInterface.register_function('rms_curvature_from_profile', rms_curvature_from_profile)
UniformTopographyInterface.register_function('rms_laplacian', rms_laplacian)
UniformTopographyInterface.register_function('rms_curvature_from_area', rms_curvature_from_area)
